#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import shutil


class filter_csv_1_0_0(ursgal.UNode):
    """filter_csv_1_0_0 UNode

    Filters .csv files row-wise according to user-defined rules.

    The filter rules have to be defined in the params. See the engine
    documentation for further information ( :meth:`.filter_csv_1_0_0._execute` ).
    """

    META_INFO = {
        'edit_version'           : 1.00,
        'name'                   : 'Filter CSV',
        'version'                : '1.0.0',
        'release_date'           : None,
        'engine_type' : {
            'misc_engine' : True
        },
        'input_extensions'       : ['.csv'],
        'output_extensions'      : ['.csv'],
        'output_suffix'          : 'accepted',
        'rejected_output_suffix' : 'rejected',
        'in_development'         : False,
        'include_in_git'         : True,
        'distributable'      : True,
        'utranslation_style'     : 'filter_csv_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'filter_csv_1_0_0.py',
                },
            },
        },
        'citation' :
        '',
    }

    def __init__(self, *args, **kwargs):
        super(filter_csv_1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        '''
        Result files (.csv) are filtered for defined filter parameters.

        Input file has to be a .csv

        Creates a _accepted.csv file and returns its path. If defined also
        rejected entries are written to _rejected.csv.

        Note:

            To write the rejected entries define 'write_unfiltered_results' as
            True in the parameters.

        Available rules:

            * lte
            * gte
            * lt
            * gt
            * contains
            * contains_not
            * equals
            * equals_not
            * regex

        Example:

            >>> params = {
            >>>     'csv_filter_rules':[
            >>>         ['PEP', 'lte', 0.01],
            >>>         ['Is decoy', 'equals', 'false']
            >>>     ]
            >>>}

        The example above would filter for posterior error probabilities lower
        than or equal to 0.01 and filter out all decoy proteins.

        Rules are defined as list of lists with the first list element as the
        column name/csv fieldname, the second list element the rule and the
        third list element the value which should be compared. Multiple rules
        can be applied, see example above. If the same fieldname should be
        filtered multiply (E.g. Sequence should not contain 'T' and 'Y'), the
        rules have to be defined separately.

        Example:

            >>> params = {
            >>>     'csv_filter_rules':[
            >>>         ['Sequence','contains_not','T'],
            >>>         ['Sequence','contains_not','Y']
            >>>     ]
            >>>}



        lte:

            'lower than or equal' (<=) value has to comparable i.e. float or
            int. Values are accepted if they are lower than or equal to the
            defined value. E.g. ['PEP','lte',0.01]

        gte:

            'greater than or equal' (>=) value has to comparable
            i.e. float or int. Values are accepted if they are greater than or
            equal to the defined value. E.g. ['Exp m/z','gte',180]

        lt:

            'lower than' (<=) value has to comparable
            i.e. float or int. Values are accepted if they are lower than
            the defined value. E.g. ['PEP','lt',0.01]

        gt:

            'greater than' (>=) value has to comparable
            i.e. float or int. Values are accepted if they are greater than
            the defined value. E.g. ['PEP','gt',0.01]

        contains:

            Substrings are checked if they are present in the the full string.
            E.g. ['Modifications','contains','Oxidation']

        contains_not:

            Substrings are checked if they are present in the the full string.
            E.g. ['Sequence','contains_not','M']

        equals:

            String comparison (==). Comparison has to be an exact match to pass.
            E.g. ['Is decoy','equals','false']. Floats and ints are not compared
            at the moment!

        equals_not:

            String comparison (!=). Comparisons differing will be rejected.
            E.g. ['Is decoy','equals_not','true']. Floats and ints are not
            compared at the moment!

        regex:

            Any regular expression matching is possible E.g. CT and CD motif
            search ['Sequence','regex','C[T|D]']


        Note:

            Some spreadsheet tools interpret False and True and show them as
            upper case when opening the files, even if they are actually
            written in lower case. This is especially important for target and
            decoy filtering, i.e. ['Is decoy','equals','false'].
            'false' has to be lower case, even if the spreadsheet tool displays
            it as 'FALSE'.

        '''
        print('[ -ENGINE- ] Executing conversion ..')
        # self.time_point(tag = 'execution')
        filter_csv_main = self.import_engine_as_python_function()
        if self.params['output_file'].lower().endswith('.csv') is False:
            raise ValueError('Trying to filter a non-csv file')

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        if self.params['translations']['write_unfiltered_results'] is False:
            output_file_unfiltered = None
        else:
            file_extension = self.META_INFO.get(
                'output_suffix',
                None
            )
            new_file_extension = self.META_INFO.get(
                'rejected_output_suffix',
                None
            )
            output_file_unfiltered = output_file.replace(
                file_extension,
                new_file_extension
            )
            shutil.copyfile(
                '{0}.u.json'.format(output_file),
                '{0}.u.json'.format(output_file_unfiltered)
            )

        filter_csv_main(
            input_file=input_file,
            output_file=output_file,
            filter_rules=self.params['translations']['csv_filter_rules'],
            output_file_unfiltered=output_file_unfiltered,
        )
        if output_file_unfiltered is not None:
            self.fix_md5_and_file_in_json(
                json_path='{0}.u.json'.format(output_file_unfiltered)
            )

        # self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/env python
import ursgal
import os
from collections import defaultdict as ddict
import csv
import sys

class pparse_2_0(ursgal.UNode):
    """
    Unode for pParse included in pGlyco 2.2.0
    For further information visit
    http://pfind.ict.ac.cn/software/pParse/#Downloads

    Note:
        Please download pParse manually as part of pGlyco 2.2.0
        https://github.com/pFindStudio/pGlyco2

    Reference:
    Yuan ZF, Liu C, Wang HP, Sun RX, Fu Y, Zhang JF, Wang LH,
    Chi H, Li Y, Xiu LY, Wang WP, He SM (2012)
    pParse: a method for accurate determination of monoisotopic peaks 
    in high-resolution mass spectra. Proteomics 12(2)
    """

    META_INFO = {
        'edit_version': 1.00,
        'name': 'pAarse',
        'version': '2.0',
        'release_date': '2018-11-02',
        'utranslation_style': 'pparse_style_1',
        'input_extensions': ['.raw'],
        'output_extensions': ['.mgf'],
        'output_suffix' : None,
        'create_own_folder': False,
        'in_development': False,
        'include_in_git': False,
        'distributable': False,
        'engine_type': {
            'converter': True,
        },
        'engine': {
            'win32' : {
                '64bit' : {
                    'exe'            : 'pParse.exe',
                    'url'            : '',
                    'zip_md5'        : None,
                    'additional_exe' : [],
                },
            },
        },
        'citation':
        'Yuan ZF, Liu C, Wang HP, Sun RX, Fu Y, Zhang JF, Wang LH,'
            'Chi H, Li Y, Xiu LY, Wang WP, He SM (2012)'
            'pParse: a method for accurate determination of monoisotopic peaks '
            'in high-resolution mass spectra. Proteomics 12(2)'
    }

    def __init__(self, *args, **kwargs):
        super(pparse_2_0, self).__init__(*args, **kwargs)
        pass

    def preflight(self):
        '''
        Formatting the command line via self.params

        Returns:
            dict: self.params

        Command line options:
            -D datapath               default (D:\data\)
            -L logfilepath            default (the same with datapath)
            -O outputpath             default (the same with datapath)
            -W isolation_width        default (2)
            -F input_format           default (raw) optional (wiff)
            -C co-elute               default (1)
            -S cut_similiar_mono      default (1)
            -I ipv_file               default (.\IPV.txt)
            -M mars_model             default (4)
            -T trainingset            default (.\TrainingSet.txt)
            -m output_mgf             default (1)
            -p output_pf              default (1)
            -d delete_msn             default (0)
            -a check_activationcenter default (1)
            -g debug_mode             default (0)
            -r rewrite_files          default (0)
            -t mars_threshold         default (-0.34)
            -u export_unchecked_mono  default (0)
            -y output_all_mars_y      default (0)
            -Y output_mars_y          default (0)
            -s output_trainingdata    default (0)
            -R recalibrate_window     default (7)
            -v outputsvmlight         default (0)
            -z m/z                    default (5)
            -i Intensity              default (1)
        '''
        self.params['translations']['raw_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        # self.created_tmp_files.append(self.param_file_name)

        self.params['command_list'] = [
            self.exe,
            '-D',
            self.params['translations']['raw_input_file'],
            # '-O',
            # self.params['translations']['output_file_incl_path']
        ]

        for flag, value in self.params['translations']['pparse_options'].items():
            self.params['command_list'].extend(
                [flag, value]
            )

        return self.params


    def postflight(self):
        '''
        Rename output file, since that naming the output file is not properly working in pParse
        '''
        os.rename(
            self.params['translations']['raw_input_file'].replace('.raw', '_HCDFT.mgf'), 
            self.params['translations']['output_file_incl_path']
        )
        return self.params['translations']['output_file_incl_path']


#!/usr/bin/env python

import ursgal
import os.path
import importlib.machinery
import sys
from pprint import pprint


class naive_bayes_1_0_0( ursgal.UNode ):
    '''
    naive_bayes_1_0_0 UNode
    '''
    META_INFO = {
        'edit_version'      : 1.00,
        'name'              : 'Naive Bayes',
        'version'           : '1.0.0',
        'release_date'      : None,
        'engine_type' : {
            'meta_engine' : True,
        },
        'input_extensions'  : ['.csv'],
        'output_extensions' : ['.csv'],
        'create_own_folder' : False,
        'include_in_git'    : True,
        'in_development'    : False,
        'distributable'      : True,
        'utranslation_style' : 'naive_bayes_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'naive_bayes_1_0_0.py',
                },
            },
        },
        'citation' : \
            'Combines PEP scores from different search engines using naive '\
            'Bayes (see i.e. http://www.paulgraham.com/naivebayes.html)',
    }

    def __init__(self, *args, **kwargs):
        super(naive_bayes_1_0_0, self).__init__(*args, **kwargs)
        pass


    def preflight( self ):
        '''
        Building the list of parameters that will be passed to the
        naive_bayes_1_0_0 main function.

        These parameters are stored in self.command_dict

        Returns:
                None
        '''
        input_file_list_for_bayes = []
        search_engine_list_for_bayes = []
        # these must have the same order!

        for input_file_dict in self.params["input_file_dicts"]:
            input_file_path = os.path.join(
                input_file_dict['dir'],
                input_file_dict['file']
            )
            input_file_list_for_bayes.append(input_file_path)
            search_engine = input_file_dict["last_engine"]
            search_engine_list_for_bayes.append(search_engine)

        features_that_define_unique_psm = \
            ['Sequence', 'Modifications', 'Spectrum Title']

        output_csv_full_path = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file'],
        )

        self.command_dict = {
            'input_csvs': input_file_list_for_bayes,
            'input_engines': search_engine_list_for_bayes,
            'output_csv': output_csv_full_path,
            'columns_for_grouping': features_that_define_unique_psm,
            'input_sep': ',',  # input csv separating char
            'output_sep': ',',  # output csv separating char
            'join_sep': ';',  # char to join multiple values in same field
            'pep_colname': 'PEP',
        }

    def _execute( self ):
        '''
        Executing the combine_FDR_0_1 main function with parameters
        that were defined in preflight (stored in self.command_dict)

        The main function is imported and then executed using the
        parameters from command_dict.

        Returns:
                None
        '''
        naive_bayes_main = self.import_engine_as_python_function()

        print('''

Executing main() function from {scriptpath} with the following parameters:

>>> {script}.main(
...     input_csvs = {input_csvs},
...     input_engines = {input_engines},
...     output_csv = '{output_csv}',
...     columns_for_grouping = {columns_for_grouping},
...     input_sep = '{input_sep}',
...     output_sep = '{output_sep}',
...     join_sep = '{join_sep}',
...     pep_colname = '{pep_colname}',
... )

        '''.format(
            scriptpath = os.path.relpath( self.exe ),
            script     = os.path.basename( self.exe ),
            **self.command_dict
            )
        )

        # executing main function of the naive bayes script
        naive_bayes_main(
            input_csvs           = self.command_dict['input_csvs'],
            input_engines        = self.command_dict['input_engines'],
            output_csv           = self.command_dict['output_csv'],
            columns_for_grouping = self.command_dict['columns_for_grouping'],
            input_sep            = self.command_dict['input_sep'],
            output_sep           = self.command_dict['output_sep'],
            join_sep             = self.command_dict['join_sep'],
            pep_colname          = self.command_dict['pep_colname'],
        )



#!/usr/bin/env python
import ursgal
import os
import subprocess
import csv


class pepnovo_3_1( ursgal.UNode ):
    """
    PepNovo v3.1 UNode
    http://proteomics.ucsd.edu/Software/PepNovo/

    Reference:
    Ari M. Frank, Mikhail M. Savitski, Michael L. Nielsen, Roman A. Zubarev, and Pavel A. Pevzner (2007) De Novo Peptide Sequencing and Identification with Precision Mass Spectrometry, J. Proteome Res. 6:114-123.
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'PepNovo',
        'version'            : 'v3.1',
        'release_date'       : None,
        'engine_type' : {
            'de_novo_search_engine' : True,
        },
        'input_extensions'   : ['.mgf'],
        'output_extensions'  : ['.csv'],
        'utranslation_style' : 'pepnovo_style_1',
        'in_development'     : False,
        'create_own_folder'  : True,
        'include_in_git'     : False,
        'distributable'      : True,
        'engine' : {
            'linux' : {
                '64bit' : {
                    'exe'            :'PepNovo_bin',
                    'url'            : '',
                    'zip_md5'        : '8f9d9125e79d7b27ffb6d60c0ccc0c97',
                    'additional_exe' : [],
                }
            },
            'darwin' : {
               '64bit' : {
                   'exe'             : 'PepNovo_bin',
               }
            },
            'win32' : {
                '64bit' : {
                    'exe'            : 'PepNovo.exe',
                }
            },
        },
        'citation' : \
            'Ari M. Frank, Mikhail M. Savitski, Michael L. Nielsen, Roman A. '\
            'Zubarev, and Pavel A. Pevzner (2007) De Novo Peptide Sequencing '\
            'and Identification with Precision Mass Spectrometry, J. Proteome '\
            'Res. 6:114-123.',
    }

    def __init__(self, *args, **kwargs):
        super(pepnovo_3_1, self).__init__(*args, **kwargs)
        self.available_mods = {
                    'Carbamidomethyl': {'fix': ('C', '+57'),
                                       'opt': ('HDE', '+57')},
                                       # 'N-term': 'NT+CAM'},
                    'Trp->Kynurenin': {'opt': ('W', '+4')},
                    'Cation:K': {'opt': ('SKNPLRVIEMDGA', '+38')},
                    'Cation:Na': {'opt': ('QVPNYTHDESFAMILG', '+22')},
                    'Acetyl': {'opt': ('LTSYGVMPDAK', '+42'),
                               'N-term': '^+42'},
                    'Carbamyl': {'opt': ('PANMETGDLFIQSVK', '+43')},
                    'Dehydrated': {'opt': ('EDST', '-18')},
                    'Dimethyl': {'opt': ('K', '+28')},
                    'Dioxidation': {'opt': ('MW', '+32')},
                    'Formyl': {'opt': ('ST', '+28')},
                    'Oxidation': {'opt': ('MHWDKNP', '+16')},
                    'Deamidated': {'opt': ('NQ', '+1')},
                    'Methyl': {'opt': ('CHKNQR', '+14'),
                               'N-term': '^+14'},
                    'Phospho': {'opt': ('STY', '+80')},
                    'Gln->pyro-Glu': {'opt': ('Q', '-17')},
                    'Amidated': {'C-term': '$-1'},
                    'Label:18O(1)': {'C-term': '$+2'},
                    }

        pass


    def preflight( self ):
        '''
        Formatting the command line via self.params

        Returns:
                dict: self.params
        '''
        self.params['translations']['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        self.params['translations']['tmp_output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file'] + '.tmp'
        )
        self.created_tmp_files.append( self.params['translations']['tmp_output_file_incl_path'])

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        if self.params['translations']['precursor_mass_tolerance_unit'] == 'ppm':
            self.params['translations']['precursor_mass_tolerance_plus'] = ursgal.ucore.convert_ppm_to_dalton(
                self.params['translations']['precursor_mass_tolerance_plus'],
                base_mz=self.params['translations']['base_mz']
            )
            self.params['translations']['precursor_mass_tolerance_minus'] = ursgal.ucore.convert_ppm_to_dalton(
                self.params['translations']['precursor_mass_tolerance_minus'],
                base_mz=self.params['translations']['base_mz']
            )
        print(
            '''
            [ WARNING ] precursor_mass_tolerance_plus and precursor_mass_tolerance_minus
            [ WARNING ] need to be combined for pepnovo_3_1 (use of symmetric tolerance window).
            [ WARNING ] The arithmetic mean is used.
            '''
        )
        self.params['translations']['precursor_mass_tolerance'] = ( float(self.params['translations']['precursor_mass_tolerance_plus']) + \
                                                    float(self.params['translations']['precursor_mass_tolerance_minus']) ) \
                                                / 2.0

        if self.params['translations']['frag_mass_tolerance_unit'] == 'ppm':
            self.params['translations']['frag_mass_tolerance'] = ursgal.ucore.convert_ppm_to_dalton(
                self.params['translations']['frag_mass_tolerance'],
                base_mz=self.params['base_mz']
            )

        if self.params['translations']['denovo_model_dir'] == None:
            self.params['translations']['denovo_model_dir'] = os.path.join(
                os.path.dirname(self.exe),
                'Models'
            )

        modifications = []
        for mod in self.params[ 'mods' ][ 'fix' ]:
            mod_available = False
            if mod['name'] in self.available_mods.keys():
                if 'fix' in self.available_mods[ mod['name'] ].keys():
                    if mod['aa'] in self.available_mods[ mod['name'] ]['fix'][0]:
                        modifications.append(mod['aa'] + self.available_mods[ mod['name'] ]['fix'][1])
                        mod_available = True
            if mod_available == False:
                print('''
                    [ WARNING ] PepNovo does not support your given modification
                    [ WARNING ] Continue without modification {0}'''.format(mod)
                    )

        for mod in self.params[ 'mods' ][ 'opt' ]:
            mod_available = False
            if mod['name'] in self.available_mods.keys():
                for term in ['N-term', 'C-term']:
                    if term in mod['pos']:
                        if term in self.available_mods[ mod['name'] ].keys():
                            modifications.append(self.available_mods[ mod['name'] ][term])
                            mod_available = True
                if 'opt' in self.available_mods[ mod['name'] ].keys():
                    if mod['aa'] in self.available_mods[ mod['name'] ]['opt'][0]:
                        modifications.append(mod['aa'] + self.available_mods[ mod['name'] ]['opt'][1])
                        mod_available = True
            if mod_available == False:
                print('''
                    [ WARNING ] PepNovo does not support your given modification
                    [ WARNING ] Continue without modification {0}'''.format(mod)
                    )

        self.params[ 'command_list' ] = [
            self.exe, # path 2 executable
            '-file', '{mgf_input_file}'.format( **self.params['translations']), # SpectrumFile (*.mzXML, *.mgf, *.ms2)
            '-model', '{denovo_model}'.format( **self.params['translations']), # (currently only CID_IT_TRYP is available)
            '-fragment_tolerance', '{frag_mass_tolerance}'.format(**self.params['translations']), # ion tolerances (in Da)
            '-pm_tolerance', '{precursor_mass_tolerance}'.format(**self.params['translations']), # precursor ion tolerance (in Da)
            '-digest', '{enzyme}'.format(**self.params['translations']),
            '-num_solutions', '{num_match_spec}'.format(**self.params['translations']),
            '-model_dir', '{denovo_model_dir}'.format(**self.params['translations']),  # - directory where model files are kept (default ./Models)
            '-PTMs', ':'.join(modifications),  # - separated by a colons (no spaces) e.g., M+16:S+80:N+1
        ]
        if self.params['translations']['pepnovo_tag_length']:
            if self.params['translations']['pepnovo_tag_length'] >= 3 and self.params['translations']['pepnovo_tag_length'] <= 6:
                self.params[ 'command_list' ].extend([
                    '-tag_length', '{pepnovo_tag_length}'.format(**self.params['translations'])
                    ]) # < 3-6> - returns peptide sequence of the specified length (only lengths 3-6 are allowed).

        translations = self.params['translations']['_grouped_by_translated_key']
        for param in [
            '-output_cum_probs',
            '-output_aa_probs',
            '-prm',
            '-prm_norm',
            '-use_spectrum_charge',
            '-use_spectrum_mz',
            '-no_quality_filter',
        ]:
            if list(translations[param].values())[0] == True:
                self.params[ 'command_list' ].append( param )

        if self.params['translations']['precursor_isotope_range'] != '0' :
            self.params[ 'command_list' ].append( '-correct_pm' )

        if self.params['translations']['min_output_score'] >= 0 and self.params['translations']['min_output_score'] <= 1.0:
            self.params[ 'command_list' ].extend([ '-min_filter_prob', '{min_output_score}'.format(**self.params['translations']) ])

        return self.params

    def _execute(self):
        if len(self.params['command_list']) != 0:
            proc = subprocess.Popen(
                self.params['command_list'],
                stdout = subprocess.PIPE,
            )
        else:
            print('Command list is empty, nothing to do here...')
            print('_execute failed ....', self.params['command_list'])
            execute_answer.append( 'Command list is empty' )
            self.execute_return_code = 500

        if proc is not None:
            output_file = open(self.params['translations']['tmp_output_file_incl_path'], 'w')
            # pint('Printing output to file, this can take a while ...')
            for line in proc.stdout:
                if line.startswith(b'>>'):
                    print(
                        'processing spectrum number: ',
                        line.strip().decode('utf').split('.')[1],
                        end = '\r'
                    )

                line_decoded = line.strip().decode('utf')
                print( line_decoded, file = output_file )
            output_file.close()

            # catching the executable's exit code to detect crashes:
            proc.communicate()[0]
            self.execute_return_code = proc.returncode
            assert self.execute_return_code in [0, None], '''
  \n{0} crashed!

  The executable
    {1}
  terminated with Error code {2} .
  Inspect the printouts above for possible causes and verify that all input files are valid.
            '''.format( self.engine, os.path.relpath(self.exe), self.execute_return_code)

        self.print_execution_time(tag='execution')
        return


    def postflight( self ):
        '''
        Reformats the PepNovo output file
        '''
        filepath = self.params['translations']['tmp_output_file_incl_path']

        print('[ PARSING  ] Loading unformatted Pepnovo results ...')

        #extract spectrum titles and spectrum ids from the pepnovo output file and also
        #save the spectrum ids in result_dict as keys.
        #Furthermore, the headers are extracted and saved into the list "headers"
        with open(filepath,'r') as pepnovo_outputfile:
            result_dict = {}
            spectrumtitle_list = []
            id_list = []
            save_headers = True
            for line in pepnovo_outputfile:
                if line.startswith('>>'):
                    line_list = line.split(' ')
                    result_dict[line_list[2]] = []
                    id_list.append(line_list[2])
                    spectrumtitle_list.append(line_list[3])
                if 'No solutions found' in line:
                    continue
                if line.startswith('#') and save_headers == True:
                    headers = line.strip('\n').split('\t')
                    save_headers = False

        #extend and translate headers
        header_translations = self.UNODE_UPARAMS['header_translations']['uvalue_style_translation']
        translated_headers = []
        for header in headers:
            if header == '':
                continue
            translated_headers.append(
                header_translations.get(header, header)
            )
        translated_headers.append('output_aa_probs')
        translated_headers.append('Spectrum Title')
        translated_headers.append('Modifications')
        translated_headers.append('Raw data location')
        translated_headers.append('Retention Time (s)')
        translated_headers.append('Calc m/z')
        translated_headers.insert(1,'Spectrum ID')

        #this section extraction from the pepnovo outputfile and stores it with the corresponding
        #spectrum ids in result_dict
        save_data = False
        spectrum_count = 0
        with open(filepath,'r') as pepnovo_outputfile:
            for line in pepnovo_outputfile.readlines():
                if line.startswith('>>'):
                    spectrum_count += 1 #just counts the number of read spectra, so that the spectrum ids can be correctly accessed in id_list
                    continue
                if line.startswith('#Index'):
                    save_data = True #specifies that the next lines should be saved in result_dict
                    result_line = []
                    continue
                if line != '\n' and save_data == True:
                    line_list = line.strip('\n').split('\t')
                    aa_probs = []
                    #this loop takes the aa probabilities from the line and puts them into a ;-separated string
                    for elem in line_list:
                        if any(c.isalpha() for c in elem) == True:
                            nr_aa = 0
                            # if spectrum_count == 1:
                            seq_index = line_list.index(elem) #specifying seq_index for the next section, but only one time
                            for char in elem:
                                if char.isalpha():
                                    nr_aa +=1
                            line_list.reverse()
                            for i in range(nr_aa):
                                aa_probs.append(line_list[0])
                                del line_list[0]
                        else:
                            continue
                    aa_probs.reverse()
                    aa_probs = ';'.join(aa_probs)
                    line_list.reverse()
                    line_list.append(aa_probs)
                    result_line.append(line_list)
                #in this part the result_line is actually saved to result_dict
                if line == '\n' and save_data == True:
                    result_dict[id_list[spectrum_count-1]] = result_line
                    save_data = False
                    continue

        print('[ Writing  ] Rewriting Pepnovo Outputfile...')

        #this section reads all sequences from result_dict, detects modifications
        #and stores them together with the corresponding aa (last_aa) and the position in moddict.
        #Afterwards, modifications get translated, deleted from sequence and stored with the old sequences as keys
        #in translated_PTMs
        translated_PTMs = {}
        for spec_id in result_dict:
            for rnk_list in result_dict[spec_id]:
                seq = rnk_list[seq_index]
                mod = ''
                moddict = {}
                seqlist = list(seq)
                for x,elem in enumerate(seq):
                    if not elem.isalpha():  #takes everything that is not a letter and adds it to mod
                        mod += elem
                        position = seqlist.index(elem)
                        del seqlist[seqlist.index(elem)] #deletes everythin that is not a letter from seqlist
                        #this counts only for the case, when a sequence ends with a modified aa
                        if seq.endswith(mod) and x == len(seq)-1:
                            if mod not in moddict:
                                moddict[mod] = (last_aa, position)
                            else:
                                #this expression facilitates the making of a list, when several of the same mods are in a seq
                                positionlist = []
                                if type(moddict[mod]) == list:
                                    positionlist = moddict[mod]
                                    positionlist.append((last_aa, position))
                                    moddict[mod] = (positionlist)
                                else:
                                    positionlist.append(moddict[mod])
                                    positionlist.append((last_aa, position))
                                    moddict[mod] = (positionlist)
                                mod = ''
                    elif mod in moddict:
                        #this expression facilitates the making of a list, when several of the same mods are in a seq
                        positionlist = []
                        if type(moddict[mod]) == list:
                            positionlist = moddict[mod]
                            positionlist.append((last_aa, position))
                            moddict[mod] = (positionlist)
                        else:
                            positionlist.append(moddict[mod])
                            positionlist.append((last_aa, position))
                            moddict[mod] = (positionlist)
                        last_aa = elem
                        mod = ''
                    elif mod == '':
                        last_aa = elem
                        continue
                    else:
                        #here, the modifications are actually stored in moddict
                        moddict[mod] = (last_aa, position)
                        last_aa = elem
                        mod = ''
                #this does still count for every rnk_list
                #in this part, the modifications are translated
                modstring = ''
                modlist = []
                for key in self.available_mods:
                    for subkey in self.available_mods[key]:
                        for mod in moddict:
                            if mod in self.available_mods[key][subkey]:
                                if type(self.available_mods[key][subkey]) == tuple:
                                    if type(moddict[mod]) == list:
                                        for elem in moddict[mod]:
                                            if elem[0] in self.available_mods[key][subkey][0]:
                                                modlist.append((str(key),int(elem[1])))
                                    elif type(moddict[mod]) == tuple:
                                        if moddict[mod][0] in self.available_mods[key][subkey][0]:
                                            modlist.append((str(key),int(moddict[mod][1])))
                                else:
                                    modlist.append((str(key),int(moddict[mod][1])))
                modlist = sorted(modlist, key=lambda x: x[1])
                for mod in modlist:
                    modstring += mod[0] + ':'+ str(mod[1]) + ';'
                #this part produces sequences without modifications
                newseq = ''
                for char in seq:
                    if char.isalpha():
                        newseq += char
                translated_PTMs[seq]= (newseq,modstring)

        #this section deletes the modifications from the aa sequences
        translated_PTMs_new = {}
        for spec_id in result_dict:
            for rnk_list in result_dict[spec_id]:
                if type(rnk_list) == str:
                    break
                for PTM in translated_PTMs.keys():
                    if PTM == rnk_list[8]:
                        rnk_list[8] = translated_PTMs[PTM][0]
                        translated_PTMs_new[translated_PTMs[PTM][0]] = (PTM,translated_PTMs[PTM][1])
        translated_PTMs = translated_PTMs_new

        #in this section the new reorganized pepnovo outputfile gets written from the collected information
        pepnovo_outputfile_new = open(self.params['translations']['output_file_incl_path'],'w')
        for header in translated_headers:
            pepnovo_outputfile_new.write(header+',')
        pepnovo_outputfile_new.write('\n')
        for x,spec_id in enumerate(id_list):
            for a in range(len(result_dict[spec_id])):
                if len(result_dict[spec_id]) == 1:
                    break
                else:
                    pepnovo_outputfile_new.write(
                    result_dict[spec_id][a][0]+','
                    +spec_id+',')
                    for b in range(1,len(result_dict[spec_id][a])):
                        pepnovo_outputfile_new.write(
                        result_dict[spec_id][a][b]+',')
                    if result_dict[spec_id][a][8] in translated_PTMs:
                        mods = translated_PTMs[result_dict[spec_id][a][8]][1]
                    pepnovo_outputfile_new.write(spectrumtitle_list[x]+',')
                    pepnovo_outputfile_new.write(mods+',')
                    pepnovo_outputfile_new.write(self.params['translations']['mgf_input_file']+',')
                    pepnovo_outputfile_new.write('False')
                    pepnovo_outputfile_new.write('\n')

        pepnovo_outputfile_new.close()

        return




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# encoding: utf-8
"""
The knowledge base holds all constants required for chemical composistion
in ursgal.

ISOTOPIC_DISTRIBUTIONS correspond to Berglund M. & Wieser M.E.
Isotopic compositions of the elements 2009 (IUPAC Technical Report)
Pure Appl. Chem., 2011, Vol. 83, No. 2, pp. 397-410
http://dx.doi.org/10.1351/PAC-REP-10-06-02
Published online 2011-01-14
http://www.ciaaw.org/pubs/TICE2009.pdf


"""
#
# ursgal
#
# Copyright (C) 2010-2011 J. Barth, A. Niehues and C. Fufezan
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

isotopic_distributions = {
    # Isotopic masses verified using
    # http://ciaaw.org/atomic-masses.htm
    # Wang, M., Audi, G., Wapstra, A. H., Kondev, F. G., MacCormick, M.,
    # Xu, X. and Pfeiffer, B. (2012)
    # The Ame2012 atomic mass evaluation. Chinese Phys. C 36, 1603–2014
    'H': [(1.0078250322, 0.999885), (2.0141017781, 0.000115)],
    'C': [(12.0000,       0.9893), (13.003354835, 0.0107)],
    'N': [(14.003074004,  0.99636), (15.000108899, 0.00364)],
    'O': [(15.994914620,  0.99757), (16.999131757, 0.00038),
          (17.999159613,  0.00205)],
    'Si': [(27.976926535,  0.92223), (28.976494665, 0.04685),
           (29.97377001,  0.03092)],
    'P': [(30.973761998,  1.00)],
    'S': [(31.972071174,  0.9499), (32.971458910, 0.0075),
          (33.9678670,  0.0425), (35.967081, 0.0001)],
    'Na': [(22.98976928,  1.00)],
    'Cl': [(34.9688527,  0.7576), (36.9659026, 0.2424)],
    'Br': [(78.918338,  0.5069), (80.916290, 0.4931)],
    'Se': [(73.9224759,  0.0089), (75.9192137, 0.0937),
           (76.9199142,  0.0763), (77.917309, 0.2377),
           (79.916522,  0.4961), (81.916700, 0.0873)]
}

aa_compositions = {
    'A': 'C3H5NO',
    'C': 'C3H5NOS',
    'D': 'C4H5NO3',
    'E': 'C5H7NO3',
    'F': 'C9H9NO',
    'G': 'C2H3NO',
    'H': 'C6H7N3O',
    'I': 'C6H11NO',
    'K': 'C6H12N2O',
    'L': 'C6H11NO',
    'M': 'C5H9NOS',
    'N': 'C4H6N2O2',
    'P': 'C5H7NO',
    'Q': 'C5H8N2O2',
    'R': 'C6H12N4O',
    'S': 'C3H5NO2',
    'T': 'C4H7NO2',
    'V': 'C5H9NO',
    'W': 'C11H10N2O',
    'Y': 'C9H9NO2',
    'U': 'C3H5NOSe'  # Selenocystein
}

monosaccharide_compositions = {
    "dHex"      : 'C6H10O4',
    "dHexNAc"   : 'C8H13NO4',
    "Hex"       : 'C6H10O5',
    "HexA"      : 'C6H8O6',
    "HexNAc"    : 'C8H13NO5',
    "Me2Hex"    : 'C8H14O5',
    "MeHex"     : 'C7H12O5',
    "MedHex"    : 'C7H12O4',
    "NeuAc"     : 'C11H17NO8',
    "Pent"      : 'C5H8O4',
    'dHexN'     : 'C6H11O3N',
    'HexN'      : 'C6H11O4N',
    'MeHexA'    : 'C7H10O6',
    'NOHCNFmPse': 'C14H22O8N2',
    'dHexNSpep' : 'C13H23O8N3S',
    'Hep'       : 'C7H12O6',
}

aa_names = {
    'A': 'alanine',
    'C': 'cysteine',
    'D': 'aspartic_acid',
    'E': 'glutamic_acid',
    'F': 'phenylalanine',
    'G': 'glycine',
    'H': 'histidine',
    'I': 'isoleucine',
    'K': 'lysine',
    'L': 'leucine',
    'M': 'methionine',
    'N': 'asparagine',
    'P': 'proline',
    'Q': 'glutamine',
    'R': 'arginine',
    'S': 'serine',
    'T': 'threonine',
    'V': 'valine',
    'W': 'tryptophan',
    'Y': 'tyrosine',
    'U': 'selenocystein'  # Selenocystein
}



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import unittest


class TestRun(unittest.TestCase):
    def setUp(self):
        self.uc = ursgal.UController()
        self.uc.params['translations'] = {}
        self.uc.UNODE_UPARAMS.update(
            {
                'test_ions' : {
                    'style'                    : 'test_style_1',
                    'ukey'                     : 'test_ions',
                    'ukey_translated'          : '__test_00000_ions',
                    'default_value'            : 'Yes',
                    'default_value_translated' : True,
                    'uvalue_style_translation' : {
                        'Yes' : True,
                        'No'  : False
                    },
                    'triggers_rerun'           : True
                },
                'score_test_ions' : {
                    'style'                    : 'test_style_1',
                    'ukey'                     : 'score_test_ions',
                    'ukey_translated'          : '__test_00000_ions',
                    'default_value'            : True,
                    'default_value_translated' : 'Please yes translate',
                    'uvalue_style_translation' : {
                        True : 'Please yes translate',
                        False: 'No please leave me alone'
                    },
                    'triggers_rerun'           : True
                },
                'list_of_things' : {
                    'style'                    : 'test_style_1',
                    'ukey'                     : 'list_of_things',
                    'ukey_translated'          : 'list_of_things',
                    'default_value'            : [ True, True, True],
                    'default_value_translated' : [ True, True, True],
                    'uvalue_style_translation' : {},
                    'triggers_rerun'           : True
                }
            }
        )

    def test_key_and_value_translation(self):
        untranslated_params, translated_params = self.uc.collect_and_translate_params(
            {
                'test_ions' : 'No',
            }
        )
        self.assertEqual(
            translated_params['test_ions_key'], '__test_00000_ions'
        )
        self.assertEqual(
            translated_params['test_ions'], False
        )


    def test_bool_translation(self):
        untranslated_params, translated_params = self.uc.collect_and_translate_params(
            {
                'score_test_ions' : False,
            }
        )
        self.assertEqual(
            translated_params['score_test_ions'], 'No please leave me alone'
        )
    def test_empty_translation_return_default(self):
        untranslated_params, translated_params = self.uc.collect_and_translate_params({})
        self.assertEqual(
            translated_params['score_test_ions'], 'Please yes translate'
        )
    def test_list_ov_values_is_kept(self):
        untranslated_params, translated_params = self.uc.collect_and_translate_params(
            {'list_of_things' : [False, False, False]}
        )
        self.assertEqual(
            translated_params['list_of_things'], [False, False, False]
        )

    def test_grouping_on_translated_keys(self):
        untranslated_params, translated_params = self.uc.collect_and_translate_params({})
        self.assertEqual(
            len(
                translated_params['_grouped_by_translated_key']['__test_00000_ions']
            ),
            2
        )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
'''
Converts mzML to dta

Usage:
./mzml2dta_0_1_0.py <mzML File1> ... <mzML FileN>

Note:
1) A new dta file will be created at the same location as the mzML file
2) only ms2 are converted to dta
'''
# 05.04.2016
# By C. Fufezan

from __future__ import print_function
import sys
import os
import argparse
import pymzml

PROTON   = 1.00727646677


def convert_mz_2_mass( mz, charge ):
    '''
    NOTE:
        equal to charge * mz - ( charge * PROTON)
    '''
    return charge * ( mz - PROTON )


def _determine_mzml_name_base( file_name ):
    file_name = os.path.basename( file_name )
    if file_name.upper().endswith('.MZML.GZ'):
        mzml_name_base = file_name[:-8]
    elif file_name.upper().endswith('.MZML'):
        mzml_name_base = file_name[:-5]
    else:
        raise Exception("Can not determine mzml base name from {0}".format(
            file_name
        ))
    return mzml_name_base


def main( mzml = None, i_decimals = 5, mz_decimals = 5 ):
    print('Converting file:\n\tmzml : {0}\n\tto\n\tdta-files'.format(
        mzml,
    ))
    mzml_name_base = _determine_mzml_name_base( mzml )
    run = pymzml.run.Reader(
        mzml,
        extraAccessions=[ ('MS:1000016', ['value', 'unitName'] )],
        obo_version = '1.1.0'
    )
    dta_entries = 0
    for n, spec in enumerate(run):
        if n % 500 == 0:
            print(
                'File : {0:^40} : Processing spectrum {1}'.format(
                    os.path.basename( mzml ),
                    n,
                ),
                end = '\r'
            )
        if spec['ms level'] != 2:
            continue
        scan_time, unit = spec['scan time']
        spectrum_id = spec['id']
        file_name = '{0}.{1}.{1}.dta'.format(
            mzml_name_base,
            spectrum_id,
        )
        with open(file_name, 'w') as oof:
            print(
                '{mh} {charge}'.format(
                    mh = convert_mz_2_mass(
                        spec['precursors'][0]['mz'],
                        spec['precursors'][0]['charge']
                    ) + PROTON,
                    charge = spec['precursors'][0]['charge']
                ),
                file=oof
            )
            for mz, intensity in spec.centroidedPeaks:
                print(
                    '{0:<10.{mzDecimals}f} {1:<10.{intensityDecimals}f}'.format(
                        mz,
                        intensity,
                        mzDecimals = mz_decimals,
                        intensityDecimals = i_decimals
                    ),
                    file = oof
                )
            print('\n', file = oof )
    print('')
    print('Wrote {0} dta entries'.format( dta_entries))
    return

if __name__ == '__main__':
    # parsing command line arguments:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'mzml',
        help='mzml input file')
    parser.add_argument(
        '-m', '--mz_decimals', default=5,
        help='Number of decimals for mz values', type=int)
    parser.add_argument(
        '-i', '--i_decimals', default=5,
        help='Number of decimals for intensity values', type=int)

    if len( sys.argv ) <= 1:
        parser.print_help()
    else:
        args = parser.parse_args()
        tmp = main( **args.__dict__ )



#!/usr/bin/env python
import ursgal
import math
import csv
import bisect
import os
from collections import defaultdict as ddict
from collections import OrderedDict
import sys

PROTON = 1.00727646677


def transform_score( score, minimum_score ):
    '''Transform scores for "bigger_is_better" == False into linear and positive space'''
    minimum_score_transformed = -1 * math.log( minimum_score, 10 )
    try:
        transformed_score = -1 * math.log( score, 10 )
    except:
        transformed_score = minimum_score_transformed

    if transformed_score >= minimum_score_transformed:
        transformed_score = minimum_score_transformed
    return transformed_score


class percolator_2_08( ursgal.UNode ):
    """
    Percolator 2_08 UNode

    q-value and posterior error probability calculation
    by a semi-supervised learning algorithm that dynamically
    learns to separate target from decoy peptide-spectrum matches (PSMs)

    Reference:
    Käll L, Canterbury JD, Weston J, Noble WS, MacCoss MJ. (2007) Semi-supervised learning for peptide identification from shotgun proteomics datasets.
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Percolator',
        'version'            : '2.08',
        'release_date'       : None,
        'engine_type' : {
            'validation_engine' : True,
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : 'percolator_2_08_validated',
        'create_own_folder'  : False,
        'include_in_git'     : False,
        'distributable'      : True,
        'group_psms'         : True,
        'in_development'     : False,
        'utranslation_style' : 'percolator_style_1',
        'engine' : {
            'darwin' : {
                '64bit' : {
                    'exe'            : 'percolator_2_08',
                    'url'            : '',
                    'zip_md5'        : '3a5f5433ec3875532e74ef54c96da3ba',
                    'additional_exe' : [],
                },
            },
            'linux' : {
                '64bit' : {
                    'exe'            : 'percolator',
                    'url'            : '',
                    'zip_md5'        : '6213411f3c56684667d56ba1a8b020ae',
                    'additional_exe' : [],
                },
            },
            'win32' : {
                '64bit' : {
                    'exe'            : 'percolator.exe',
                    'url'            : '',
                    'zip_md5'        : '60bc939789ef83e11cde4e75cb4ca9e2',
                    'additional_exe' : [],
                },
                # '32bit' : {
                #     'exe'            : 'percolator.exe',
                #     'url'            : '',
                #     'zip_md5'        : '8b1387860c15d07e938a517ddd2cfffa',
                #     'additional_exe' : [],
                # },
            },
        },
        'citation' : \
            'Käll L, Canterbury JD, Weston J, Noble WS, MacCoss MJ. (2007) '\
            'Semi-supervised learning for peptide identification from shotgun '\
            'proteomics datasets.',
    }

    def __init__(self, *args, **kwargs):
        super(percolator_2_08, self).__init__(*args, **kwargs)
        pass

    def preflight( self ):
        '''
        Formating the command line to via self.params
        '''

        PERCOLATOR_FIELDS = OrderedDict([
            (
                'SpecId', {
                    'csv_field': 'Spectrum Title',
                    'DefaultDirection': 'DefaultDirection'
                }
            ),
            (
                'Label', {
                    'csv_field': '',
                    'DefaultDirection': '-'
                }
            ),
            (
                'ScanNr', {
                    'csv_field': 'Spectrum ID',
                    'DefaultDirection': '-'
                }
            ),
            (
                'lnrSp', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                'The natural logarithm of the rank of the match based on the Sp score'
                }
            ),
            (
                'deltLCn', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                "The difference between this PSM's XCorr and the XCorr of the last-ranked \
                PSM for this spectrum, divided by this PSM's XCorr or 1, whichever is larger."
                }
            ),
            (
                'deltCn', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                    "The difference between this PSM's XCorr and the XCorr of the next-ranked \
                PSM for this spectrum, divided by this PSM's XCorr or 1, whichever is larger. \
                Note that this definition differs from that of the standard delta Cn reported \
                by SEQUEST・ゑｽｮ"
                        }
            ),
            (
                'Xcorr', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                        "The SEQUEST cross-correlation score"
                        }
            ),
            (
                'Sp', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                "The preliminary SEQUEST score."
                        }
            ),
            (
                'IonFrac', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                "The fraction of b and y ions theoretical ions matched to the spectrum"
                        }
            ),
            (
                'Mass',    {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': \
                "The observed mass [M+H]+"
                        }
            ),
            # (
            #     'PepLen',  {
            #         'csv_field': '',
            #         'DefaultDirection': 0,
            #         'description': \
            #     "The length of the matched peptide, in residues"
            #     }
            # ),
            (
                'Charge1', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge2', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge3', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge4', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge5', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge6', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge7', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge8', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge9', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'Charge10', {
                    'csv_field': '',
                    'DefaultDirection': 0
                }
            ),
            (
                'enzN', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': "Is the peptide preceded by an enzymatic (tryptic) site?"
                }
            ),
            (
                'enzC', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': "Does the peptide have an enzymatic (tryptic) C-terminus?"
                }
            ),
            (
                'enzInt', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': "Number of missed internal enzymatic (tryptic) sites"
                }
            ),
            (
                'lnNumSP', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': "The natural logarithm of the number of database peptides within the \
                    specified precursor range"
                }
            ),
            (
                'dM', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': "The difference between the calculated and observed mass"
                }
            ),
            (
                'absdM', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': "The absolute value of the difference between the calculated and \
                    observed mass"
                }
            ),
            (
                'Peptide', {
                    'csv_field': '',
                    'DefaultDirection': 0,
                    'description': ""
                }
            ),
            (
                'Proteins', {
                    'csv_field': 'Protein ID',
                    'DefaultDirection': 0,
                    'description': ""
                }
            ),
        ])


        self.params['translations']['csv_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        # this file will contain the decoys only (we need it for fdr calculations and such)
        self.params['translations']['decoy_output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            "decoysOnly_" + self.params['output_file']
        )

        self.params['translations']['percolator_in'] = \
            '{output_file_incl_path}.tsv'.format(
                **self.params['translations']
        )

        self.params['translations']['percolator_out'] = \
            '{output_file_incl_path}.psms'.format(**self.params['translations'])

        # writing the Percolator input file (tab separated format)
        o = open( self.params['translations']['percolator_in'], 'w')
        writer = csv.DictWriter(
            o,
            list(PERCOLATOR_FIELDS.keys()),
            delimiter='\t',
            extrasaction='ignore'
        )
        writer.writeheader()

        self.params['translations']['percolator_decoy_out'] = \
            '{decoy_output_file_incl_path}.psms'.format(**self.params['translations'])
        self.params['command_list'] = [
            # percolator -X pout.xml pin.tab >| yeast-01.psms
            self.exe,
            '--only-psms',
            '{percolator_in}'.format(**self.params['translations']),
            '--results-psms',
            '{percolator_out}'.format(**self.params['translations']),
            '--decoy-results-psms',
            '{percolator_decoy_out}'.format(**self.params['translations'])
        ]

        last_search_engine = self.get_last_search_engine(
            history = self.stats['history']
        )
        self.params['_score_list'] = self.generating_score_list()
        minimum_score = None

        bigger_scores_better = self.UNODE_UPARAMS['bigger_scores_better']['uvalue_style_translation'][last_search_engine]

        if bigger_scores_better is False:
            for p, _score in enumerate(self.params['_score_list']):
                if _score <= 0:
                    # jumping over truly zero or negative values ...
                    continue
                else:
                    s = -1 * math.log( _score, 10 )
                    if minimum_score is None:
                        minimum_score = _score
                # OMSSA HACK OMG PLEASE HELP US ALL
                if "OMSSA" in last_search_engine.upper():
                    fo = transform_score(_score, minimum_score)
                    if fo < 100:
                        minimum_score = _score
                        break
                else:
                    break

        for n, spectrum_title in enumerate( self.params['grouped_psms'].keys()):
            best_score = self.params['grouped_psms'][ spectrum_title ][0][0]
            worst_score = self.params['grouped_psms'][ spectrum_title ][ -1 ][0]

            if bigger_scores_better is False:
                best_score = transform_score(best_score, minimum_score)
                worst_score = transform_score(worst_score, minimum_score)

            for m, (score, line_dict) in enumerate(
                    self.params['grouped_psms'][ spectrum_title ]):
                t = {}

                if bigger_scores_better is True:
                    rank_of_score = bisect.bisect_right(
                        self.params['_score_list'],
                        score
                    )
                else:
                    rank_of_score = bisect.bisect_left(
                        self.params['_score_list'],
                        score
                    )
                    rank_of_score = len( self.params['_score_list'] ) - rank_of_score
                #
                if 'Mascot:Rank' in line_dict.keys():
                    rank_of_score = int(line_dict['Mascot:Rank'])
                    # if best_score <= 5:
                    #     continue
                t['lnrSp'] = math.log( 1 + rank_of_score )
                t['Sp'] = rank_of_score
                #
                charge      = float(line_dict['Charge'])
                exp_mz      = float(line_dict['Exp m/z'])
                t['Mass']   = ( exp_mz * charge ) - ( charge - 1 ) * PROTON
                #
                t['Xcorr'] = score
                if bigger_scores_better is False:
                    t['Xcorr'] = transform_score(t['Xcorr'], minimum_score)

                t['PepLen'] = len( line_dict['Sequence'] )
                t['Charge{Charge}'.format(**line_dict)] = 1

                normalization = t['Xcorr']
                if t['Xcorr'] < 1:
                    normalization = 1
                if m == len( self.params['grouped_psms'][ spectrum_title ] ) - 1:
                    # last entry
                    deltLCn = 0
                    deltCn = 0
                else:
                    deltLCn = (t['Xcorr'] - worst_score) / normalization

                    next_score = self.params['grouped_psms'][ spectrum_title ][m + 1][0]
                    if bigger_scores_better is False:
                        next_score = transform_score(next_score, minimum_score)

                    deltCn = (t['Xcorr'] - next_score ) / normalization
                t['deltCn'] = deltCn
                t['deltLCn'] = deltLCn
                if line_dict['Is decoy'].upper() == 'TRUE':
                    t['Label'] = -1
                else:
                    t['Label'] = 1

                # if self.params['translations']['decoy_tag'] in line_dict['proteinacc_start_stop_pre_post_;']:
                #     # bug mzIdentML msgf+ convert
                #     t['Label'] = -1

                # this - is - sparta (or, if you like mzIdentML) ...
                # splitted = line_dict['proteinacc_start_stop_pre_post_;'].split('_')
                # aka http://imgur.com/WjiX9
                pre_aa = []
                for prot_pre_aa in line_dict['Sequence Pre AA'].split(self.params['translations']['protein_delimiter']):
                    for p_pre_aa in prot_pre_aa.split(';'):
                        pre_aa.append(p_pre_aa)
                post_aa = []
                for prot_post_aa in line_dict['Sequence Post AA'].split(self.params['translations']['protein_delimiter']):
                    for p_post_aa in prot_post_aa.split(';'):
                        post_aa.append(p_post_aa)
                allowed_aa = set(self.params['translations']['enzyme'].split(';')[0] + '-')
                cleavage_site = self.params['translations']['enzyme'].split(';')[1]
                inhibitor_aa = self.params['translations']['enzyme'].split(';')[2]
                final_pre_aa = pre_aa[0]
                final_post_aa = post_aa[0]
                t['enzN'] = 0
                t['enzC'] = 0
                if cleavage_site == 'C':
                    for i, aa in enumerate(pre_aa):
                        if aa in allowed_aa  \
                            or line_dict['Sequence Start'] in ['1', '2']:
                            t['enzN'] = 1
                            final_pre_aa = aa
                            final_post_aa = post_aa[i]
                        if line_dict['Sequence'][-1] in allowed_aa\
                            or '-' in post_aa:
                            t['enzC'] = 1
                elif cleavage_site == 'N':
                    for i, aa in enumerate(post_aa):
                        if aa in allowed_aa:
                            t['enzC'] = 1
                            final_post_aa = aa
                            final_pre_aa = pre_aa[i]
                        if line_dict['Sequence'][0] in allowed_aa\
                            or line_dict['Sequence Start'] in ['1','2']:
                            t['enzN'] = 1

                t['enzInt'] = 0
                if self.params['translations']['enzyme'] != 'ACDEFGHIKLMNPQRSTVWY;C;':
                    for aa in line_dict['Sequence'][:-1]:
                        if aa in allowed_aa:
                            t['enzInt'] += 1

                t['dM'] = float(line_dict['Calc m/z']) - float(line_dict['Exp m/z'])
                t['absdM'] = abs(t['dM'])

                mods = line_dict['Modifications']
                if mods.strip() == '':
                    t['Peptide'] = '{0}.{Sequence}.{1}'.format(
                        sorted(pre_aa)[0],
                        sorted(post_aa)[0],
                        **line_dict
                    )
                else:
                    t['Peptide'] = '{0}.{Sequence}#{1}.{2}'.format(
                        sorted(pre_aa)[0],
                        mods,
                        sorted(post_aa)[0],
                        **line_dict
                    )
                # although peptides without mod ill have # at their end,
                # mapping further down will still work ...

                for per_key in PERCOLATOR_FIELDS.keys():
                    mapped_key = PERCOLATOR_FIELDS[ per_key ]['csv_field']
                    if 'Charge' in per_key:
                        if per_key not in t.keys():
                            t[ per_key ] = 0
                    if mapped_key != '':
                        t[ per_key ] = line_dict[ mapped_key ].strip()
                writer.writerow( t )
        o.close()


        # marking temporary files for deletion:
        self.created_tmp_files += [
            self.params['translations']['decoy_output_file_incl_path'],
            self.params['translations']['percolator_in'],
            '{output_file_incl_path}.psms'.format( **self.params['translations'] ),
            '{output_file_incl_path}.peptides'.format( **self.params['translations'] ),
        ]


    def postflight( self ):
        '''
        read the output and merge in back to the ident csv
        '''
        percolator_log_file_name  = self.params['translations']['output_file_incl_path'].replace(
            '.csv',
            '_percolator_log.txt'
        )
        percolator_log_file = open(
            percolator_log_file_name,
            'w'
        )
        potential_buggy_percolator_output = self.params['translations']['percolator_out'] + '.psms'
        if os.path.exists( potential_buggy_percolator_output ):
            print('WTF Percolator ?')
            print('Renaming: \n{percolator_out}.psms ->> {percolator_out}'.format(
                **self.params['translations']
            ))
            os.rename(
                self.params['translations']['output_file_incl_path'] + '.psms.psms',
                self.params['translations']['output_file_incl_path'] + '.psms'
            )
            os.rename(
                self.params['translations']['decoy_output_file_incl_path'] + '.psms.psms',
                self.params['translations']['decoy_output_file_incl_path'] + '.psms'
            )
            os.rename(
                self.params['translations']['output_file_incl_path'] + '.psms.peptides',
                self.params['translations']['output_file_incl_path'] + '.peptides'
            )

        s2l = {
            'target': ddict(list),
            'decoy' : ddict(list)
        }
        for pkey, p_out in [('target','percolator_out'), ('decoy','percolator_decoy_out')]:

            percolator_output_dict_reader = csv.DictReader(
                open(
                    self.params['translations'][ p_out ],
                    'r'
                ),
                delimiter='\t'
            )
            for line_dict in percolator_output_dict_reader:

                peptide = line_dict['peptide'].split('.')[1]
                psmid_pep_key = (
                    line_dict['PSMId'],
                    peptide,
                )
                if psmid_pep_key not in s2l[ pkey ].keys():
                    s2l[ pkey ][ psmid_pep_key ] = line_dict


        opened_file = open( self.params['translations'][ 'csv_input_file' ], 'r' )
        csv_input = csv.DictReader( row for row in opened_file if not row.startswith('#') )

        if "PEP" not in csv_input.fieldnames and "q-value" not in csv_input.fieldnames:
            csv_input.fieldnames += ['PEP', 'q-value']
        csv_kwargs = {}

        if sys.platform == 'win32':
            csv_kwargs['lineterminator'] = '\n'
        else:
            csv_kwargs['lineterminator'] = '\r\n'

        percolator_stats = ddict(int)
        csv_output = csv.DictWriter(
            open( self.params['translations']['output_file_incl_path'], 'w' ),
            csv_input.fieldnames,
            **csv_kwargs
        )

        csv_output.writeheader()
        for line_dict in csv_input:

            # check if the current line is a decoy or a target
            # so we know in which percolator output file we have to look for it.
            psm_type = "target"
            if line_dict['Is decoy'].upper() == 'TRUE':
                psm_type = "decoy"
            # if self.params['translations']['decoy_tag'] in line_dict['proteinacc_start_stop_pre_post_;']:
            #     line_dict['Is decoy'] = "true"
            #     psm_type = "decoy"
            if line_dict['Modifications'].strip() != '':
                seq_and_mods = '#'.join( [line_dict['Sequence'], line_dict['Modifications']] )
            else:
                seq_and_mods = line_dict['Sequence']
            _psmid_pep_key = (
                line_dict['Spectrum Title'],
                seq_and_mods,
            )
            if _psmid_pep_key in s2l[ psm_type ].keys():
                line_dict['PEP']     = s2l[ psm_type ][ _psmid_pep_key ]['posterior_error_prob']
                line_dict['q-value'] = s2l[ psm_type ][ _psmid_pep_key ]['q-value']
                # write all results including decoy into the full csv:
                csv_output.writerow( line_dict )
                percolator_stats['PSMs accepted and kept by Percolator'] += 1
            else:
                print(
                    'Original PSM :{0} could not be found in percolator output file, most probably because PSM was filtered by percolator, (multiple peptides to one spectrum match)'.format(
                        _psmid_pep_key
                    ),
                    file = percolator_log_file
                )
                percolator_stats['PSMs filtered and removed by Percolator'] += 1
        for title, count in sorted(percolator_stats.items()):
            self.print_info(
                '{0} : {1}'.format(
                    title,
                    count
                ),
                caller = 'Info'

            )
        self.print_info(
            'Please refer to file {0} for details'.format(percolator_log_file_name),
            caller = 'Info'
        )

    def generating_score_list( self ):
        scores = []
        for k, v in self.params['grouped_psms'].items():
            for score, line_dict in v:
                scores.append( score )
        # this will depnd on the scoring scheme ...
        # e.g. OMSSA smaller better - XTandem bigger better
        scores.sort()  # reverse=params['bigger_scores_better'])
        return scores



#!/usr/bin/env python
import ursgal
import os
from collections import defaultdict as ddict
import csv
import itertools
import sys

class pglyco_fdr_2_2_0(ursgal.UNode):
    """
    Unode for pGlycoFDR included in pGlyco 2.2.0
    This node allows post-processing of pGlyco results

    Note:
        Please download pGlycoFDR manually as part of pGlyco 2.2.0
        https://github.com/pFindStudio/pGlyco2

    Reference:
    Liu MQ, Zeng WF,, Fang P, Cao WQ, Liu C, Yan GQ, Zhang Y, Peng C, Wu JQ,
    Zhang XJ, Tu HJ, Chi H, Sun RX, Cao Y, Dong MQ, Jiang BY, Huang JM, Shen HL,
    Wong CCL, He SM, Yang PY. (2017) pGlyco 2.0 enables precision N-glycoproteomics
    with comprehensive quality control and one-step mass spectrometry
    for intact glycopeptide identification.
    Nat Commun 8(1)
    """

    META_INFO = {
        'edit_version': 1.00,
        'name': 'pGlycoFDR',
        'version': '2.2.0',
        'release_date': '2019-01-01',
        'utranslation_style': 'pglyco_fdr_style_1',
        'input_extensions': ['.csv'],
        'output_extensions': ['.csv'],
        'create_own_folder': False,
        'in_development': False,
        'include_in_git': False,
        'distributable': False,
        'engine_type': {
            'validation_engine': True,
        },
        'engine': {
            'win32' : {
                '64bit' : {
                    'exe'            : 'pGlycoFDR.exe',
                    'url'            : '',
                    'zip_md5'        : None,
                    'additional_exe' : [],
                },
            },
        },
        'citation':
        'Liu MQ, Zeng WF,, Fang P, Cao WQ, Liu C, Yan GQ, Zhang Y, Peng C, Wu JQ,'
            'Zhang XJ, Tu HJ, Chi H, Sun RX, Cao Y, Dong MQ, Jiang BY, Huang JM, Shen HL,'
            'Wong CCL, He SM, Yang PY. (2017) pGlyco 2.0 enables precision N-glycoproteomics '
            'with comprehensive quality control and one-step mass spectrometry'
            'for intact glycopeptide identification.'
            'Nat Commun 8(1)'
    }

    def __init__(self, *args, **kwargs):
        super(pglyco_fdr_2_2_0, self).__init__(*args, **kwargs)
        pass

    def preflight(self):
        '''
        Formatting the command line and writing the param input file via 
        self.params

        Returns:
            dict: self.params
        '''
        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        search_result_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        print(search_result_file)

        self.param_file_name = search_result_file.strip('.csv') + '_pGlyco.cfg'

        self.params['command_list'] = [
            self.exe,
            '-p',
            self.param_file_name,
            '-r',
            os.path.join(
                self.params['input_dir_path'],
                'pGlycoDB-GP.txt',
            ),
        ]

        return self.params

    def postflight(self):
        # '''
        # Reads pGlycoFDR txt output and write final csv output file.

        # Adds:
        #     * Raw data location, since this can not be added later

        # '''
        pglyco_header = [
            'GlySpec',
            'PepSpec',
            'RawName',
            'Scan',
            'RT',
            'PrecursorMH',
            'PrecursorMZ',
            'Charge',
            'Rank',
            'Peptide',
            'Mod',
            'PeptideMH',
            'Glycan(H,N,A,G,F)',
            'PlausibleStruct',
            'GlyID',
            'GlyFrag',
            'GlyMass',
            'GlySite',
            'TotalScore',
            'PepScore',
            'GlyScore',
            'CoreMatched',
            'CoreFuc',
            'MassDeviation',
            'PPM',
            'GlyIonRatio',
            'PepIonRatio',
            'GlyDecoy',
            'PepDecoy',
            'GlycanFDR',
            'PeptideFDR',
            'TotalFDR',
        ]

        translated_headers = []
        header_translations = self.UNODE_UPARAMS[
            'header_translations']['uvalue_style_translation']
        # import pprint
        # pprint.pprint(self.meta_unodes['pglyco_fdr_2_2_0'].UNODE_UPARAMS)
        for original_header_key in pglyco_header:
            ursgal_header_key = header_translations[original_header_key]
            translated_headers.append(ursgal_header_key)

        csv_kwargs = {}
        if sys.platform == 'win32':
            csv_kwargs['lineterminator'] = '\n'
        else:
            csv_kwargs['lineterminator'] = '\r\n'
        csv_writer = csv.DictWriter(
            open(self.params['translations']['output_file_incl_path'], 'w'),
            fieldnames=translated_headers,
            **csv_kwargs
        )

        pglyco_output = os.path.join(
            self.params['output_dir_path'],
            'pGlycoDB-GP-FDR.txt'
        )
        csv_reader = csv.DictReader(
            open(pglyco_output, 'r'),
            fieldnames=translated_headers,
            delimiter='\t',
        )
        # self.created_tmp_files.append(pglyco_output)

        csv_writer.writeheader()
        for n, line_dict in enumerate(csv_reader):
            if n == 0:
                continue
            csv_writer.writerow(line_dict)
        return



#!/usr/bin/env python
'''
usage:
    ./upeptide_mapper_1_0_0.py <input_file> <output_file>


'''

from __future__ import print_function
import sys
import os
import csv
import ursgal
import re
from collections import Counter, defaultdict
import itertools
from copy import deepcopy
try:
    import regex as regex
    finditer_kwargs = { 'overlapped' : True }
except:
    print('[ WARNING  ] Standard re module cannot find overlapping pattern')
    print('[   INFO   ] Consider installing the regex module')
    print('[   INFO   ] pip install -r requirements.txt')
    import re as regex
    finditer_kwargs = {}
import bisect
# increase the field size limit to avoid crash if protein merge tags
# become too long does not work under windows
if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)



def main(input_file=None, output_file=None, params=None):
    '''
    Peptide mapping implementation as Unode.

    Arguments:
        input_file (str): input filename of csv
        output_file (str): output filename
        params (dict): dictionary containing ursgal params


    Results and fixes
        * All peptide Sequences are remapped to their corresponding protein,
          assuring correct start, stop, pre and post aminoacid.
        * It is determined if the corresponding proteins are decoy proteins.
          These peptides are reported after the mapping process.
        * Non-mappable peptides are reported. This can e.g. due to 'X' in
          protein sequences in the fasta file or other non-standard amino acids.
          These are sometimes replaced/interpreted/interpolated by the search
          engine.
          A recheck is performed if the peptides can be mapped containing an 'X'
          at any position. These peptides are also reported. If peptides can
          still not be mapped after re-mapping, these are reported as well.


    '''
    print(
        '''[ map_peps ] Mapping peptides from file {0}'''.format(
            os.path.basename(input_file),
        )
    )
    do_not_delete      = False
    created_tmp_files  = []
    target_decoy_peps  = set()
    non_mappable_peps  = set()
    pep_map_lookup     = {}
    joinchar           = params['translations']['protein_delimiter']
    if sys.platform == 'win32':
        try:
            import ahocorasick
        except:
            print(
'[ WARNING ] pyahocorasick can not be installed via pip on Windows at the moment\n'
'[ WARNING ] Falling back to UpeptideMapper_v2'
            )
            params['translations']['peptide_mapper_class_version'] = 'UPeptideMapper_v2'

    if params['translations']['peptide_mapper_class_version'] == 'UPeptideMapper_v2':
        upapa = UPeptideMapper_v2( word_len = params['translations']['word_len'])
        fasta_lookup_name = upapa.build_lookup_from_file(
            params['translations']['database'],
            force  = False,
        )
        class_etxra_args = [fasta_lookup_name]
    elif params['translations']['peptide_mapper_class_version'] == 'UPeptideMapper_v3':
        #silent default :)
        upapa = UPeptideMapper_v3( params['translations']['database'] )
        fasta_lookup_name = upapa.fasta_name
        class_etxra_args = [ fasta_lookup_name ]
    elif params['translations']['peptide_mapper_class_version'] == 'UPeptideMapper_v4':
        #silent default :)
        upapa = UPeptideMapper_v4( params['translations']['database'] )
        fasta_lookup_name = os.path.basename( params['translations']['database'] )
        class_etxra_args = []

    else:
        print(
            '[ map_peps ] peptide mapper class version unknown: {0}'.format(
                params['translations']['peptide_mapper_class_version']
            )
        )
        sys.exit(1)
    print(
        '''[ map_peps ] Using peptide mapper version: {0}'''.format(
            params['translations']['peptide_mapper_class_version']
        )
    )

    output_file_object = open(output_file, 'w')
    protein_id_output  = open(output_file + '_full_protein_names.txt', 'w')
    mz_buffer = {}
    csv_kwargs = {
        'extrasaction' : 'ignore'
    }
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'

    pep_map_lookup = {}

    with open( input_file, 'r' ) as in_file:
        csv_input  = csv.DictReader(
            in_file
        )

        output_fieldnames = list(csv_input.fieldnames)
        for remove_fieldname in [
            'Start',
            'Stop',
            'gi',
            'Accession',
            'proteinacc_start_stop_pre_post_;'

        ]:
            if remove_fieldname not in output_fieldnames:
                continue
            output_fieldnames.remove(remove_fieldname)
        new_fieldnames = [
            'Is decoy',
            'Protein ID',
            'Sequence Start',
            'Sequence Stop',
            'Sequence Pre AA',
            'Sequence Post AA',
        ]

        for new_fieldname in new_fieldnames:
            if new_fieldname not in output_fieldnames:
                output_fieldnames.insert(-5,new_fieldname)
        csv_output = csv.DictWriter(
            output_file_object,
            output_fieldnames,
            **csv_kwargs
        )
        csv_output.writeheader()
        print('''[ map_peps ] Parsing and buffering csv''')
        import time
        csv_file_buffer = []
        tmp_peptide_set = set()
        for line_dict in csv_input:
            # we should add here all the J to I and L variants of MSAmanda...
            appended = False
            for aa_to_replace, replace_dict in sorted(params['translations']['aa_exception_dict'].items(), reverse=True):
                if aa_to_replace in line_dict['Sequence']:
                    if aa_to_replace in ['O']:
                        if 'unimod_name' in replace_dict.keys():
                            for r_pos, aa in enumerate(line_dict['Sequence']):
                                if aa == aa_to_replace:
                                    index_of_aa = r_pos + 1
                                    unimod_name = replace_dict['unimod_name']
                                    new_mod = '{0}:{1}'.format(
                                        unimod_name,
                                        index_of_aa
                                    )
                                    if line_dict['Modifications'] == '':
                                        line_dict['Modifications'] += new_mod
                                    else:
                                        line_dict['Modifications'] += ';{0}'.format(
                                            new_mod
                                        )
                        if len(replace_dict['original_aa']) == 1:
                            line_dict['Sequence'] = line_dict['Sequence'].replace(
                                aa_to_replace,
                                replace_dict['original_aa'][0]
                            )
                    elif aa_to_replace in ['J']:
                        repeat_count = line_dict['Sequence'].count(aa_to_replace)
                        #we need to do all combos for the positions of I and L in the seq...
                        positions = []
                        for index_of_aa, aa in enumerate(line_dict['Sequence']):
                            if aa == aa_to_replace:
                                positions.append( index_of_aa )
                        product_combos = []
                        new_peptides = set()
                        for combo_tuple in itertools.product(replace_dict['original_aa'],repeat=repeat_count):
                            # product_combos.append(n)
                            for pos, new_aa in enumerate(combo_tuple):
                                index_of_aa = positions[pos]
                                # text[:1] + 'Z' + text[2:]
                                remapped_peptide = '{0}{1}{2}'.format(
                                    line_dict['Sequence'][:index_of_aa],
                                    new_aa,
                                    line_dict['Sequence'][index_of_aa+1:],
                                )
                                tmp_peptide_set.add(
                                    remapped_peptide
                                )
                                line_dict['Sequence'] = remapped_peptide
                                csv_file_buffer.append( deepcopy(line_dict) )
                                appended = True
                    else:
                        print('''
[ WARNING ] New not covered case of aa exception for: "{0}"
[ WARNING ] Please adjust upeptide_mapper accordingly'''.format(aa_to_replace)
                        )
                        sys.exit(1)
            if appended is False:
                csv_file_buffer.append(line_dict)
            tmp_peptide_set.add( line_dict['Sequence'] )
        num_peptides = len(tmp_peptide_set)
        print('''[ map_peps ] Parsing and buffering csv done''')
        print(
            '''[ map_peps ] Mapping {0} peptides using mapper {1}'''.format(
                num_peptides,
                params['translations']['peptide_mapper_class_version']
            )
        )
        p2p_mappings  = upapa.map_peptides(
            list(tmp_peptide_set),
            *class_etxra_args
        )
        print('''[ map_peps ] Mapping peptides done''')
        total_lines = len(csv_file_buffer)
        #this assertion works for all but MSAmanda sinc J instaead of I or L is reported
        # print(set(tmp_peptide_set)-set(p2p_mappings.keys()))
        # print(len(set(tmp_peptide_set)-set(p2p_mappings.keys())))
        # assert len(p2p_mappings.keys()) == len(tmp_peptide_set)
        for line_nr, line_dict in enumerate(csv_file_buffer):
            if line_nr % 500 == 0:
                print(
                    '[ map_peps ] Processing line number: {0}/{1} .. '.format(
                        line_nr,
                        total_lines,
                    ),
                    end='\r'
                )
            # remap peptides to proteins, check correct enzymatic
            # cleavage and decoy assignment
            lookup_identifier = '{0}><{1}'.format(
                line_dict['Sequence'],
                fasta_lookup_name
            )
            if lookup_identifier not in pep_map_lookup.keys():
                tmp_decoy = set()
                upeptide_maps = p2p_mappings[line_dict['Sequence']]

                if upeptide_maps  == []:
                    print('''
[ WARNING ] The peptide {0} could not be mapped to the
[ WARNING ] given database {1}
[ WARNING ] {2}
[ WARNING ] This PSM will be skipped.
                        '''.format(
                            line_dict['Sequence'],
                            fasta_lookup_name,
                            ''
                        )
                    )
                    non_mappable_peps.add(line_dict['Sequence'])
                    continue

                sorted_upeptide_maps = [ protein_dict for protein_dict in sorted( upeptide_maps, key=lambda x: x['id'] ) ]

                protein_mapping_dict = None
                last_protein_id      = None
                for protein in sorted_upeptide_maps:
                    if protein_mapping_dict is None:
                        protein_mapping_dict = {
                            'Protein ID'       : protein['id'],
                            'Sequence Start'   : str(protein['start']),
                            'Sequence Stop'    : str(protein['end']),
                            'Sequence Pre AA'  : protein['pre'],
                            'Sequence Post AA' : protein['post'],
                        }
                    else:
                        if protein['id'] == last_protein_id:
                            tmp_join_char = ';'
                        else:
                            tmp_join_char = joinchar
                            protein_mapping_dict['Protein ID' ] += '{0}{1}'.format(tmp_join_char, protein['id'])

                        protein_mapping_dict['Sequence Start'   ] += '{0}{1}'.format(tmp_join_char, str(protein['start']))
                        protein_mapping_dict['Sequence Stop'    ] += '{0}{1}'.format(tmp_join_char, str(protein['end']))
                        protein_mapping_dict['Sequence Pre AA'  ] += '{0}{1}'.format(tmp_join_char, protein['pre'])
                        protein_mapping_dict['Sequence Post AA' ] += '{0}{1}'.format(tmp_join_char, protein['post'])

                    last_protein_id = protein['id']

                    # mzidentml-lib does not always set 'Is decoy' correctly
                    # (it's always 'false' for MS-GF+ results), this is fixed here:
                    if params['translations']['decoy_tag'] in protein['id']:
                        tmp_decoy.add('true')
                    else:
                        tmp_decoy.add('false')

                if len(protein_mapping_dict['Protein ID']) >= 2000:
                    print(
                        '{0}: {1}'.format(
                            line_dict['Sequence'],
                            protein_mapping_dict['Protein ID']
                        ),
                        file = protein_id_output
                    )
                    protein_mapping_dict['Protein ID'] = protein_mapping_dict['Protein ID'][:1990] + ' ...'
                    do_not_delete = True

                if len(tmp_decoy) >= 2:
                    target_decoy_peps.add(line_dict['Sequence'])
                    protein_mapping_dict['Is decoy'] = 'true'
                else:
                    protein_mapping_dict['Is decoy'] = list(tmp_decoy)[0]

                pep_map_lookup[ lookup_identifier ] = protein_mapping_dict

            line_dict.update( pep_map_lookup[lookup_identifier] )

            csv_output.writerow(line_dict)
    output_file_object.close()

    if len(target_decoy_peps) != 0:
        print(
'''[ WARNING ] {0}
[ WARNING ] These {1} peptides above (truncated to 100) occurred in a target as well as decoy protein
[ WARNING ] 'Is decoy' has been set to 'True' '''.format(
    target_decoy_peps if len(target_decoy_peps) <100 else list(target_decoy_peps)[:99],
    len(target_decoy_peps)
)
        )
    if len(non_mappable_peps) != 0:
        print(
'''[ WARNING ] {0}
[ WARNING ] These {1} peptides above (truncated to 100) could not be mapped to the database
[ WARNING ] Check Search and Database if neccesary'''.format(
    sorted(list(non_mappable_peps)) if len(list(non_mappable_peps)) < 100 else list(non_mappable_peps)[:99],
    len(non_mappable_peps)
)
        )
        #recheck these peptides if sequence has a 'X'
        # this is done by the peptide regex function in the unode... Use this instead?
        print('[ map_peps ] Attempting re-map of non-mappable peptides')
        peptide_has_X_in_sequence = set()
        mappable_after_all = set()
        for non_mappable_peptide in non_mappable_peps:
            if non_mappable_peptide == '':
                continue
            variants  = set()
            for pos, aa in enumerate(non_mappable_peptide):
                variants.add(
                    '{0}{1}{2}'.format(
                        non_mappable_peptide[:pos],
                        'X',
                        non_mappable_peptide[pos+1:],
                    )

                )
            # reset buffer...
            upapa.peptide_2_protein_mappings[fasta_lookup_name] = defaultdict(list)
            p2p_mappings  = upapa.map_peptides(
                list(variants),
                *class_etxra_args
            )
            if len(p2p_mappings.keys()) != 0:
                for p_with_x, maps in p2p_mappings.items():
                    if maps != []:
                        peptide_has_X_in_sequence.add( p_with_x )
                        mappable_after_all.add( non_mappable_peptide )
        print(
'''[ WARNING ] {0}
[ WARNING ] These {1} not mappable peptides (truncated to 100) have "X" in their sequence
[ WARNING ] {2} are part of the non-mappable peptides'''.format(
    sorted(list(peptide_has_X_in_sequence)) if len(peptide_has_X_in_sequence) < 100 else sorted(list(peptide_has_X_in_sequence))[:99],
    len(peptide_has_X_in_sequence),
    len(mappable_after_all)
)
        )
        not_mappable_after_all = non_mappable_peps - mappable_after_all
        if len(not_mappable_after_all) != 0:
            print(
'''[ WARNING ] {0}
[ WARNING ] These {1} peptides (truncated to 100) are indeed not mappable
[ WARNING ] Check of Search parameters and database is strongly recommended'''.format(
    not_mappable_after_all if len(not_mappable_after_all) < 100 else list(not_mappable_after_all)[:99],
    len(not_mappable_after_all),
)
            )

    if do_not_delete is False:
        created_tmp_files.append( output_file + '_full_protein_names.txt' )
    return created_tmp_files


class UPeptideMapper_v2( dict ):
    '''
    UPeptideMapper class offers ultra fast peptide to sequence mapping using
    a fast cache, hereafter referred to fcache.

    The fcache is build using the `build_lookup_from_file` or `build_lookup`
    functions. The fcache can be queried using the UPeptideMapper.map_peptide()
    function.

    Note:

        This is the deprectaed version of the peptide mapper which can be used
        by setting the parameter 'peptide_mapper_class_version' to 'UPeptideMapper_v2'.
        Otherwise the new mapper class version ('UPeptideMapper_v3') is used as default.

    '''
    def __init__(self, word_len = None ):
        self.fasta_sequences = {}
        self.word_len = word_len
        self.hits = {'fcache': 0, 'regex': 0}
        self.query_length = defaultdict(int)
        self.master_buffer = {}
        self.peptide_2_protein_mappings = {}
        pass

    def build_lookup_from_file( self, path_to_fasta_file, force=True):
        '''
        Builds the fast cache and regular sequence dict from a fasta stream

        return the internal fasta name, i.e. dirs stripped away from the path
        '''
        abs_path = os.path.abspath( path_to_fasta_file )
        assert os.path.exists( abs_path ), '''
            file {0} not found'''.format( abs_path )

        internal_name = os.path.basename(abs_path)
        with open( abs_path, 'r') as io:
            self.build_lookup(
                fasta_name   = internal_name,
                fasta_stream = io.readlines(),
                force        = force,
            )
        return internal_name

    def build_lookup( self, fasta_name=None, fasta_stream=None, force=True ):
        '''
        Builds the fast cache and regular sequence dict from a fasta stream
        '''
        print('[ upapa v2 ] UPeptideMapper is building the fast cache and regular sequence dict from a fasta')
        if fasta_name not in self.keys():
            force = True
            self[ fasta_name ] = {}
            self.fasta_sequences[ fasta_name ] = {}

        self.master_buffer[ fasta_name ] = {}
        # self.word_len = 6
        if force:
            for upapa_id, (id, seq) in enumerate(
                    ursgal.ucore.parse_fasta( fasta_stream )):
                print(
                    '[ upapa v2 ] Indexing sequence #{0} with word_len {1}'.format(
                        upapa_id,
                        self.word_len
                    ),
                    end = '\r'
                )
                if seq.endswith('*'):
                    seq = seq.strip('*')
                self.fasta_sequences[ fasta_name ][ id ] = seq
                self._create_fcache(
                    id         = id,
                    seq        = seq,
                    fasta_name = fasta_name,
                )
                # if upapa_id > 10000:
                #     break
            print()
        # input('bp')
        # exit(1)

    def _create_fcache(self, id=None, seq=None, fasta_name=None):
        '''
        Updates the fast cache with a given sequence
        '''
        for pos in range(len( seq ) - self.word_len + 1):
            pep = seq[ pos : pos + self.word_len ]
            # sorted 3.8GB
            s_pep = ''.join(sorted(pep))
            try:
                self[ fasta_name ][ s_pep ].add( (id, pos + 1) )
            except:
                self[ fasta_name ][ s_pep ] = set([ (id, pos + 1) ])

        return

    def map_peptides(self, peptide_list, fasta_name = None, force_regex = False ):
        '''
        Wrapper function to map a given peptide list in one batch.

        Args:
            peptide_list (list): list with peptides to be mapped
            fasta_name (str): name of the database


        '''
        if fasta_name not in self.peptide_2_protein_mappings.keys():
            self.peptide_2_protein_mappings[fasta_name] = defaultdict(list)
        for peptide in peptide_list:
            self.peptide_2_protein_mappings[fasta_name][peptide] = self.map_peptide(
                peptide,
                fasta_name = fasta_name,
            )

        return self.peptide_2_protein_mappings[fasta_name]

    def map_peptide(self, peptide=None, fasta_name=None, force_regex=False):
        '''
        Maps a peptide to a fasta database.

        Returns a list of single hits which look for example like this::

            {
                'start' : 12,
                'end'   : 18,
                'id'    : 'Protein Id passed to the function',
                'pre'   : 'A',
                'post'  : 'V',
            }
        '''

        l_peptide                        = len(peptide)
        mappings                         = []
        required_hits                    = l_peptide - self.word_len

        self.query_length[ l_peptide ] += 1
        if fasta_name in self.keys():
            if peptide in self.master_buffer[fasta_name].keys():
                mappings = self.master_buffer[fasta_name][peptide]
            else:
                if l_peptide < self.word_len or force_regex:
                    self.hits['regex'] += 1
                    pattern = regex.compile( r'''{0}'''.format( peptide ))
                    # we have to through it by hand ...
                    # for fasta
                    for id, seq in self.fasta_sequences[ fasta_name ].items():
                        for match in pattern.finditer( seq, **finditer_kwargs ):
                            start = match.start() + 1
                            end   = match.end()
                            hit = self._format_hit_dict(  seq, start, end, id )
                            mappings.append( hit )
                else:
                    self.hits['fcache'] += 1
                    tmp_hits = {}
                    # m = []
                    for pos in range( l_peptide - self.word_len + 1):
                        pep = peptide[ pos : pos + self.word_len ]
                        # pep = pep.encode()
                        # print( pep, peptide )
                        s_pep = ''.join(sorted(pep))

                        fasta_set = self[ fasta_name ].get(s_pep, None)
                        if fasta_set is None:
                            continue

                        for id, id_pos in fasta_set:
                            try:
                                tmp_hits[ id ].add( id_pos )
                            except:
                                tmp_hits[ id ] = set([id_pos ])

                    for id, pos_set in tmp_hits.items():
                        sorted_positions = sorted(pos_set)
                        seq = self.fasta_sequences[ fasta_name ][ id ]

                        for n, pos in enumerate( sorted_positions ):
                            start = sorted_positions[n]
                            end   = sorted_positions[n] + l_peptide - 1

                            if n + required_hits >= len( sorted_positions):
                                break

                            expected_number = pos + required_hits
                            try:
                                observed_number = sorted_positions[ n + required_hits ]
                            except:
                                break

                            if expected_number == observed_number or required_hits == 0:
                                if seq[ start - 1 : end ] == peptide:
                                    # double check
                                    mappings.append(
                                        self._format_hit_dict(
                                            seq, start, end, id
                                        )
                                    )
                self.master_buffer[fasta_name][peptide] = mappings
        return mappings

    def _format_hit_dict( self, seq, start, end, id ):
        '''
        Creates a formated dictionary from a single mapping hit. At the same
        time evaluating pre and pos amino acids from the given sequence
        Final output looks for example like this::

            {
                'start' : 12,
                'end'   : 18,
                'id'    : 'Protein Id passed to the function',
                'pre'   : 'A',
                'post'  : 'V',
            }

        Note:
            If the pre or post amino acids are N- or C-terminal, respectively,
            then the reported amino acid will be '-'

        '''
        if start == 1:
            pre_aa = '-'
        else:
            pre_aa = seq[ start - 2]
        if end >= len(seq):
            post_aa = '-'
        else:
            post_aa = seq[ end ]
        hit = {
            'start' : start,
            'end'   : end,
            'id'    : id,
            'pre'   : pre_aa ,
            'post'  : post_aa,
        }
        return hit

    def purge_fasta_info( self, fasta_name ):
        '''
        Purges regular sequence lookup and fcache for a given fasta_name
        '''
        del self.fasta_sequences[ fasta_name ]
        del self[ fasta_name ]



class UPeptideMapper_v3():
    '''
    UPeptideMapper V3

    New improved version which is faster and consumes less memory than earlier
    versions. Is the new default version for peptide mapping.

    Note:
        Uses the implementation of Aho-Corasick algorithm pyahocorasick.
        Please refer to https://pypi.python.org/pypi/pyahocorasick/ for more
        information.

    Warning:
        The new implementation is still in beta/testing phase. Please use, check
        and interpret accordingly



    '''
    def __init__(self, fasta_database ):
        self.fasta_name                 = os.path.basename(os.path.abspath( fasta_database ))

        self.protein_indices            = defaultdict(dict)
        self.protein_sequences          = defaultdict(dict)

        self.total_sequence_list        = defaultdict(list)
        self.protein_list               = defaultdict(list)

        self.fasta_counter              = defaultdict(int)
        self.len_total_sequence_string  = defaultdict(int)

        self.peptide_2_protein_mappings = {}
        self.total_sequence_string      = {}
        self.cache_database(fasta_database, self.fasta_name)

        self.automatons = {}


    def cache_database(self, fasta_database,  fasta_name):
        '''
        Function to cache the given fasta database.

        Args:
            fasta_database (str): path to the fasta database
            fasta_name (str): name of the database
                (e.g. os.path.basename(fasta_database))

        Note:

            If the same fasta_name is buffered again all info is purged from the
            class.
        '''
        if fasta_name in self.protein_indices.keys():
            self.purge_fasta_info(fasta_name)
            self.fasta_name = fasta_name
        for protein_pos, (protein_id, seq) in enumerate(ursgal.ucore.parse_fasta(open(fasta_database,'r').readlines())):
            if protein_pos % 5000 == 0:
                print(
                    '[ upapa v3 ] Buffering protein #{0} of database {1}'.format(
                        self.fasta_counter[fasta_name],
                        os.path.basename(fasta_database)
                    ),
                    end ='\r'
                )
            len_seq             = len(seq)

            self.protein_indices[fasta_name][protein_id] = {
                'start': self.len_total_sequence_string[fasta_name],
                'stop' : self.len_total_sequence_string[fasta_name] + len_seq
            }
            # self.total_sequence_string += seq
            self.total_sequence_list[fasta_name].append(seq)
            self.protein_list[fasta_name] += [ protein_id ] * len_seq
            self.protein_sequences[fasta_name][ protein_id ] = seq
            self.len_total_sequence_string[fasta_name] += len_seq
            self.fasta_counter[fasta_name] += 1
        print()
        # print('[ upapa v3 ] Joining protein sequences')
        self.total_sequence_string[fasta_name] = ''.join( self.total_sequence_list[fasta_name] )
        # print('[ upapa v3 ] Joining protein sequences done')
        return

    def map_peptides(self, peptide_list, fasta_name):
        '''
        Function to map a given peptide list in one batch.

        Args:
            peptide_list (list): list with peptides to be mapped
            fasta_name (str): name of the database
                (e.g. os.path.basename(fasta_database))

        Returns:
            peptide_2_protein_mappings (dict): Dictionary containing
                peptides as keys and lists of protein mappings as values of the
                given fasta_name

        Note:
            Based on the number of peptides the returned mapping dictionary
            can become very large.

        Warning:
            The peptide to protein mapping is resetted if a new list o peptides
            is mapped to the same database (fasta_name).

        Examples::

            peptide_2_protein_mappings['BSA1']['PEPTIDE']  = [
                {
                    'start' : 1,
                    'end'   : 10,
                    'pre'   : 'K',
                    'post'  : 'D',
                    'id'    : 'BSA'
                }
            ]
        '''
        import ahocorasick
        # if fasta_name not in self.peptide_2_protein_mappings.keys():
        self.peptide_2_protein_mappings[fasta_name] = defaultdict(list)

        # self.peptide_2_protein_mappings = defaultdict(list)
        self.automatons[fasta_name] = ahocorasick.Automaton()
        for idx, peptide in enumerate(peptide_list):
            #integrated buffering of peptides
            if peptide not in self.peptide_2_protein_mappings[fasta_name].keys():
                self.automatons[fasta_name].add_word(peptide, (idx, peptide))
        self.automatons[fasta_name].make_automaton()
        for match in self.automatons[fasta_name].iter(self.total_sequence_string[fasta_name]):
            idx, (p_idx, m_peptide) = match
            len_m_peptide = len(m_peptide)
            protein_name_end_index = self.protein_list[fasta_name][idx]
            protein_name_start_index = self.protein_list[fasta_name][idx-len_m_peptide+1]
            if protein_name_end_index != protein_name_start_index:
                #overlap between end and start of next sequence!!!
                # print(m_peptide)
                # print(protein_name_start_index)
                # print(protein_name_end_index)
                # sys.exit(1)
                continue
            protein_start_in_sequence_string = self.protein_indices[fasta_name][protein_name_end_index]['start']
            protein_stop_in_sequence_string  = self.protein_indices[fasta_name][protein_name_end_index]['stop']

            protein_seq = self.protein_sequences[fasta_name][protein_name_end_index]
            # protein_seq = self.total_sequence_string[protein_start_in_sequence_string:protein_stop_in_sequence_string]
            # assert protein_seq == self.protein_sequences[fasta_name][protein_name_end_index]
            stop_in_protein = idx - protein_start_in_sequence_string + 1
            start_in_protein = stop_in_protein - len(m_peptide)
            # print(m_peptide, start_in_protein, stop_in_protein)
            # print(protein_seq[start_in_protein: stop_in_protein])
            assert m_peptide == protein_seq[start_in_protein: stop_in_protein]
            pre_pos = start_in_protein - 1
            if pre_pos < 0:
                pre = '-'
            else:
                pre = protein_seq[ pre_pos ]
            try:
                post = protein_seq[ stop_in_protein ]
            except:
                post = '-'

            self.peptide_2_protein_mappings[fasta_name][m_peptide].append(
                {
                    'start' : start_in_protein + 1,
                    'end'   : stop_in_protein,
                    'pre'   : pre,
                    'post'  : post,
                    'id'    : protein_name_end_index
                }
            )

        return self.peptide_2_protein_mappings[fasta_name]

    def purge_fasta_info( self, fasta_name ):
        '''
        Purges regular sequence lookup and fcache for a given fasta_name
        '''
        print('[ upapa v3 ] Purging buffer for {0}'.format(fasta_name))
        del self.protein_list[fasta_name]
        del self.protein_indices[fasta_name]
        del self.protein_sequences[fasta_name]
        del self.total_sequence_string[fasta_name]
        del self.fasta_counter[fasta_name]
        del self.total_sequence_list[fasta_name]
        del self.len_total_sequence_string[fasta_name]
        if fasta_name in self.peptide_2_protein_mappings.keys():
            del self.peptide_2_protein_mappings[fasta_name]
        if fasta_name in self.automatons.keys():
            del self.automatons[fasta_name]


class UPeptideMapper_v4():
    '''
    UPeptideMapper V4

    Improved version of class version 3 (changes proposed by Christian)

    Note:
        Uses the implementation of Aho-Corasick algorithm pyahocorasick.
        Please refer to https://pypi.python.org/pypi/pyahocorasick/ for more
        information.



    '''
    def __init__(self, fasta_database ):

        # self.protein_indices            = {}
        self.protein_sequences          = {}

        self.total_sequence_list        = []

        self.protein_list               = []
        self.protein_start_indices      = []

        self.fasta_counter              = 0
        self.len_total_sequence_string  = 0

        self.peptide_2_protein_mappings = {}
        self.total_sequence_string      = {}
        self.cache_database(fasta_database)


    def cache_database(self, fasta_database):
        '''
        Function to cache the given fasta database.

        Args:
            fasta_database (str): path to the fasta database

        Note:

            If the same fasta_name is buffered again all info is purged from the
            class.
        '''

        for protein_pos, (protein_id, seq) in enumerate(ursgal.ucore.parse_fasta(open(fasta_database,'r').readlines())):
            if protein_pos % 5000 == 0:
                print(
                    '[ upapa v4 ] Buffering protein #{0} of database {1}'.format(
                        self.fasta_counter,
                        os.path.basename(fasta_database)
                    ),
                    end ='\r'
                )
            len_seq             = len(seq)

            self.total_sequence_list.append(seq)

            self.protein_list.append( protein_id )
            self.protein_start_indices.append( self.len_total_sequence_string +protein_pos) # protein start

            self.protein_sequences[ protein_id ] = seq
            self.len_total_sequence_string += len_seq  #offset for delimiter |
            self.fasta_counter += 1
        print()
        self.total_sequence_string = '|'.join( self.total_sequence_list )
        return

    def map_peptides(self, peptide_list):
        '''
        Function to map a given peptide list in one batch.

        Args:
            peptide_list (list): list with peptides to be mapped


        Returns:
            peptide_2_protein_mappings (dict): Dictionary containing
                peptides as keys and lists of protein mappings as values of the
                given fasta_name

        Note:
            Based on the number of peptides the returned mapping dictionary
            can become very large.

        Warning:
            The peptide to protein mapping is resetted if a new list o peptides
            is mapped to the same database (fasta_name).

        Examples::

            peptide_2_protein_mappings['PEPTIDE']  = [
                {
                    'start' : 1,
                    'end'   : 10,
                    'pre'   : 'K',
                    'post'  : 'D',
                    'id'    : 'BSA'
                }
            ]
        '''
        import ahocorasick
        self.peptide_2_protein_mappings = defaultdict(list)

        # self.peptide_2_protein_mappings = defaultdict(list)
        self.automaton =  ahocorasick.Automaton()
        for idx, peptide in enumerate(peptide_list):
            self.automaton.add_word(peptide, (idx, peptide))
        self.automaton.make_automaton()

        for match in self.automaton.iter(self.total_sequence_string):
            idx, (p_idx, m_peptide) = match
            len_m_peptide = len(m_peptide)

            protein_index = bisect.bisect( self.protein_start_indices, idx ) - 1
            protein_name  = self.protein_list[protein_index]
            protein_seq   = self.protein_sequences[protein_name]
            start_in_total_seq_string = self.protein_start_indices[protein_index]
            # print('Index',protein_index)
            stop_in_protein = idx - start_in_total_seq_string + 1
            start_in_protein = stop_in_protein - len(m_peptide)

            pre_pos = start_in_protein - 1

            # print(stop_in_protein)
            # print(start_in_protein)
            # print(pre_pos)
            if pre_pos < 0:
                pre = '-'
            else:
                pre = protein_seq[ pre_pos ]
            try:
                post = protein_seq[ stop_in_protein ]
            except:
                post = '-'

            self.peptide_2_protein_mappings[m_peptide].append(
                {
                    'start' : start_in_protein + 1,
                    'end'   : stop_in_protein,
                    'pre'   : pre,
                    'post'  : post,
                    'id'    : protein_name
                }
            )

        return self.peptide_2_protein_mappings




if __name__ == '__main__':
    if len(sys.argv) < 5:
        print(__doc__)
        sys.exit(1)

    params = {
        'translations' : {
            'modifications' : [
                'M,opt,any,Oxidation',        # Met oxidation
                'C,fix,any,Carbamidomethyl',  # Carbamidomethylation
                '*,opt,Prot-N-term,Acetyl',    # N-Acteylation[]
            ],
            'aa_exception_dict' : {
                'J' : {
                    'original_aa' : ['I', 'L'],
                },
                'O' : {
                    'original_aa' : ['K'],
                    'unimod_name' : 'Methylpyrroline',
                },
            },
            'protein_delimiter'        : '<|>',
            'decoy_tag'                : 'decoy_',
            'word_len'                 : 6
        },
        'prefix' : None
    }
    params['translations']['database']                     = sys.argv[3]
    params['translations']['peptide_mapper_class_version'] = sys.argv[4]

    main(
        input_file     = sys.argv[1],
        output_file    = sys.argv[2],
        params         = params,
    )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
R = ursgal.UController()

TESTS = [
    {
        'input'  : [{'1': 0, '2': 0 }, {'1': 1, '2': 0 }],
        'output' : {'1': [0, 1], '2': [0, 0]}
    },
]


def merge_fdicts_test():
    for test_id, test_dict in enumerate(TESTS):
        yield merge_fdicts, test_dict


def merge_fdicts( test_dict ):
    output = R.merge_fdicts( *test_dict['input'] )
    print( output , test_dict)
    assert output == test_dict['output'], '''
        merge_fdicts {0} failed with output {1}'''.format(
        test_dict,
        output
    )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        merge_fdicts( test_dict )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import sys
import shutil


def main(mzML_file=None, profile=None, database=None):
    '''
    Executes a search with XTandem on an mzml_input_file.
    Results are validated with percolator and exported to .ssl.

    usage:
        ./search_and_conversion_to_ssl.py <mzML_file> <profile> <target_decoy_database>

    '''
    uc = ursgal.UController(
        profile=profile,
        params={
            'database': database,
            'modifications': [
                'M,opt,any,Oxidation',        # Met oxidation
                'C,fix,any,Carbamidomethyl',  # Carbamidomethylation
                '*,opt,Prot-N-term,Acetyl'    # N-Acteylation
            ],
        }
    )

    if sys.maxsize > 2 ** 32:
        xtandem = 'xtandem_vengeance'
    else:
        xtandem = 'xtandem_sledgehammer'

    unified_search_result_file = uc.search(
        input_file=mzML_file,
        engine=xtandem,
        force=False
    )

    validated_results = uc.validate(
        input_file=unified_search_result_file,
        engine='percolator_2_08',
    )

    uc.params['ssl_score_column_name'] = 'q-value'
    uc.params['ssl_score_type'] = 'PERCOLATOR QVALUE'

    ssl_file = uc.convert(
        input_file=validated_results,
        engine='csv2ssl',
    )
    return


if __name__ == '__main__':
    main(
        mzML_file=sys.argv[1],
        profile=sys.argv[2],
        database=sys.argv[3]
    )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import sys

R = ursgal.UController()

TESTS = [
    {
        'input' : os.path.join('tests', 'data', 'test.json'),
        'output' : {
            'unix' : '379450895e2c116886b2e92dfcd68b2b',
            'win32': '54c19ed069413037dc857a0130dd2527'
        }
    },
    {
        'input' : os.path.join('tests', 'data', 'test_without_database.json'),
        'output' : {
            'unix' : 'deb20d01ff369188a583decf203cf769',
            'win32': '26cc3e0850d3ab95c74e4cf680475335'
        }
    }
]


def check_md5_test():
    for test_id, test_dict in enumerate(TESTS):
        yield check_md5, test_dict


def check_md5( test_dict ):
    out_put = R.calc_md5( test_dict['input'] )
    # print( out_put , test_dict)
    # platform = sys.platform
    # if sys.platform != 'win32':
    #     platform = 'unix'

    assert out_put in [test_dict['output']['unix']]+[test_dict['output']['win32']], '''
        MD5 {0} failed
        output: {1}'''.format(
            test_dict,
            out_put
        )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        check_md5( test_dict )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import unittest
import pprint
from ursgal import umapmaster as umama


TEST_DICT = {
    'enzyme' : {
        'default_value' : 'trypsin',
        'available_in_unode' : ['omssa', 'xtandem'],
        'description' : 'Enzyme used for digestions',
        'ukey_translation' : {
            'omssa_style_1' : '-ot',
            'xtandem_style_1' : 'enzyme'
        },
        'uvalue_translation' : {
            'omssa_style_1' : {
                'trypsin' : 1
            },
            'xtandem_style_1' : {
                'trypsin' : '[KR]|{P}'
            }
        }
    },
    'frag_mass_type' : {
        'default_value' : 'monoisotopic',
        'available_in_unode' : ['omssa'],
        'description' : 'use chemical average or monoisotopic mass for fragment ions.',
        'ukey_translation' : {
            'omssa_style_1' : '-tom',
        },
        'uvalue_translation' : {
            'omssa_style' : {
                'monoisotopic' : 0,
                'average' : 1
            }
        }
    },
}


class UMapMaster(unittest.TestCase):
    def setUp(self):
        self.umama_small = umama.UParamMapper( TEST_DICT )
        self.umama_default = umama.UParamMapper()

    def test_all_mappings_by_style(self):
        results = []
        for mDict in self.umama_small.mapping_dicts('omssa'):
            results.append( (mDict['ukey'], mDict['default_value']) )
        self.assertEqual(
            sorted(results),
            sorted([
                ('enzyme', 'trypsin'),
                ('frag_mass_type', 'monoisotopic' )
            ])
        )

    def test_get_masked_params( self ):
        mask = ['default_value', 'description']
        masked_params = self.umama_small.get_masked_params(mask=mask)
        self.assertEqual(
            len(masked_params.keys()),
            len(TEST_DICT.keys())
        )
        self.assertEqual(
             len(masked_params['enzyme'].keys()), len(mask)
        )
        self.assertEqual(
            list(sorted(masked_params['enzyme'].keys())), sorted(mask)
        )




if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Converts xTandem.xml files into .csv
We need to do this on our own, because mzidentml_lib
reports wrong positions for modifications
(and it is also not able to convert the piledriver.mzid into csv)

It should be noted that
- xtandem groups are not merged (since it is not the same as protein groups)
- multiple domains (multiple occurence of a peptide in the same protein) are not reported

(C)
Christian Fufezan
Stefan Schulze
"""
from __future__ import print_function
import sys, os
import csv
import math
from xml.etree import cElementTree
from collections import defaultdict as ddict
import codecs

if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)

def main(input_file = None, decoy_tag = None, output_file = None):
    '''
    Converts xTandem.xml files into .csv
    We need to do this on our own, because mzidentml_lib
    reports wrong positions for modifications
    (and it is also not able to convert the piledriver.mzid into csv)

    It should be noted that
    - xtandem groups are not merged (since it is not the same as protein groups)
    - multiple domains (multiple occurence of a peptide in the same protein) are not reported

    '''
    NEW_HEADERS = [
        'Raw data location',
        'Spectrum ID',
        'Spectrum Title',
        'Retention Time (s)',
        'Rank',
        'Calc m/z',
        'Exp m/z',
        'Charge',
        'Sequence',
        'Modifications',
        'X\!Tandem:expect',
        'X\!Tandem:hyperscore',
        'proteinacc_start_stop_pre_post_;',
        'Is decoy',
    ]
    PROTON = 1.00727646677
    protein = None
    group = None
    group_counter = 0
    protein_groups = []

    if input_file.endswith('.gz'):
        compressed = True
        # Gzipped files are not seekable
        import gzip
        # file_object = codecs.getreader("utf-8")(
        #     gzip.open(path)
        # )
        file_object = codecs.getreader("utf-8")(
            gzip.open(input_file)
        )
    else:
        file_object = codecs.open(
            input_file,
            mode     = 'r',
            encoding = 'utf-8'
        )

    csvOut     = csv.DictWriter( open( output_file , 'w', newline='') , NEW_HEADERS )
    csvOut.writeheader()

    print("Converting XTandem XML into CSV: {0}".format(input_file))
    tandemXML = iter(cElementTree.iterparse( file_object, events = ( b'start',b'end')))

    for pos, (event, element) in enumerate(tandemXML):
        if event == b'start':
            if element.tag.endswith('bioml'):
                raw_data_location = element.attrib['label'].split("'")[1]
            if element.tag.endswith('group'):
                if 'mh' in element.attrib.keys():
                    group_counter = 0
                    notes                         = ddict(set)
                    notes_key                     = None
                    group                         = {} # reset group
                    group['Raw data location']    = raw_data_location
                    group['Charge']               = element.attrib['z']
                    group['Exp m/z']              = (float(element.attrib['mh'])/float(group['Charge'])) + PROTON
                    group['X\!Tandem:expect']     = element.attrib['expect']
                    protein_groups             = []
                    domain_groups             = []
                if element.attrib['label'] == 'fragment ion mass spectrum':
                    notes_key = 'spectrum'
                if element.attrib['label'] == 'input parameters':
                    notes_key = 'parameters'
                group_counter += 1

            elif element.tag.endswith('protein'):
                protein = {} # reset protein
                notes_key = 'protein'


            elif element.tag.endswith('domain'):
                if 'seq' in element.attrib.keys():
                    domain = { header: '' for header in NEW_HEADERS } # reset domain
                    domain['Sequence']              = element.attrib['seq']
                    domain['X\!Tandem:hyperscore']  = element.attrib['hyperscore']
                    domain['Calc m/z'] = (float(element.attrib['mh'])/float(group['Charge'])) + PROTON
                    protein['Start']      = element.attrib['start']
                    protein['Stop']       = element.attrib['end']
                    protein['pre']        = element.attrib['pre'][-1]
                    protein['post']       = element.attrib['post'][0]
                    if protein['Start'] == '1':
                        protein['pre'] = '-'
                    if protein['post'] == ']':
                        protein['post'] = '-'
                    domain['proteinacc_start_stop_pre_post_;'] = '{0}_{1}_{2}_{3}_{4};'.format(
                        protein['Defline'],
                        protein['Start'],
                        protein['Stop'],
                        protein['pre'],
                        protein['post'],
                    )
                    if protein['Defline'].startswith(decoy_tag):
                        domain['Is decoy'] = True
                    else:
                        domain['Is decoy'] = False
                    domain['Rank'] = 1

            elif element.tag.endswith('aa'):
                if 'modified' not in element.attrib.keys():
                    continue
                mod = '{0}:{1}'.format(element.attrib['modified'], int(element.attrib['at']) - int(protein['Start']) + 1)
                try:
                    domain['Modifications'].append( mod )
                except:
                    domain['Modifications'] =  [ mod ]
        else:
            if element.tag.endswith('protein'):
                protein_groups.append( protein )
            if element.tag.endswith('domain'):
                for k,v in domain.items():
                    if k not in protein.keys():
                        protein[k] = v
                # domain_groups.append( domain )
            elif element.tag.endswith('note'):
                if notes_key == 'spectrum':
                    group['Spectrum Title'] = element.text.strip()
                elif notes_key == 'protein':
                    protein['Defline'] = element.text.strip()
            elif element.tag.endswith('group'):
                group_counter -= 1
                if group_counter == 0:
                    dict2write = {}
                    # proteins = []
                    for protein_group in protein_groups:
                        for key in NEW_HEADERS:
                            # if key == 'proteinacc_start_stop_pre_post_;':
                            #     proteins.append(protein_group[key])
                            #     continue
                            if key == 'Modifications' and protein_group[key] !='':
                                dict2write[key] = ';'.join(protein_group[key])
                                continue
                            try:
                                dict2write[key] = group[key]
                            except:
                                dict2write[key] = protein_group[key]
                        # dict2write['proteinacc_start_stop_pre_post_;'] = ''.join(proteins)
                        csvOut.writerow( dict2write )
                    protein_groups = []
    return

if __name__ == '__main__':
    script = main(input_file = sys.argv[1], decoy_tag = sys.argv[2], output_file = sys.argv[3])



#!/usr/bin/env python
import ursgal
import os


class xtandem_sledgehammer( ursgal.UNode ):
    """
    X!Tandem UNode
    Parameter options at http://www.thegpm.org/TANDEM/api/

    Reference:
    Craig R, Beavis RC. (2004) TANDEM: matching proteins with tandem mass spectra.
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'X!Tandem',
        'version'                     : 'Sledgehammer',
        'release_date'                : '2013-9-1',
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'input_extensions'            : ['.mgf', '.gaml', '.dta', '.pkl', '.mzData', '.mzXML'],
        'output_extensions'           : ['.xml'],
        'create_own_folder'           : True,
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'               : True,
        'utranslation_style'          : 'xtandem_style_1',
        'engine' : {
            'darwin' : {
                '64bit' : {
                    'exe'            : 'tandem',
                    'url'            : '',
                    'zip_md5'        : 'e153fbd41a51ce8eeb08ca12caa418f4',
                    'additional_exe' : [],
                },
            },
            'linux' : {
                '64bit' : {
                    'exe'            : 'tandem',
                    'url'            : '',
                    'zip_md5'        : 'dfe9e4d850aeb34bfb256b027615544b',
                    'additional_exe' : [],
                },
                # 'i386' : {
                #     'exe'            : 'tandem',
                #     'url'            : '',
                #     'zip_md5'        : '9b11bcdf86137ffdcee08e8940053e19',
                #     'additional_exe' : [],
                # },
            },
            'win32' : {
                '64bit' : {
                    'exe'            : 'tandem.exe',
                    'url'            : '',
                    'zip_md5'        : '2c5e45292a36500f83c1a22dd62c0bda',
                    'additional_exe' : [],
                },
                # '32bit' : {
                #     'exe'            : 'tandem.exe',
                #     'url'            : '',
                #     'zip_md5'        : '86b78ea31389ab7d2634d0272ea6371f',
                #     'additional_exe' : [],
                # },
            },
        },
        'citation' : 'Craig R, Beavis RC. (2004) TANDEM: matching proteins '\
        'with tandem mass spectra.',
    }

    def __init__(self, *args, **kwargs):
        super(xtandem_sledgehammer, self).__init__(*args, **kwargs)
        pass

    def preflight( self ):
        '''
        Formatting the command line via self.params

        Input files from format_templates are created in the output folder
        and added to self.created_tmp_files (can be deleted)

        Returns:
            dict: self.params
        '''

        self.params['translations']['mgf_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        xml_required = [
            'default_input.xml',
            'taxonomy.xml',
            '15N-masses.xml',
            'input.xml'
        ]
        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        for file_name in xml_required:
            file_info_key = file_name.replace('.xml','')
            xml_file_path = os.path.join(
                self.params['output_dir_path'],
                file_name
            )
            self.params['translations'][ file_info_key ] = xml_file_path
            self.created_tmp_files.append( xml_file_path )
        #
        # building command_list !
        #
        self.params['command_list'] =[
            self.exe,
            '{input}'.format(**self.params['translations']),
        ]

        if self.params['translations']['label'] == '15N':
            self.params['translations']['15N_default_input_addon'] = '<note label="protein, modified residue mass file" type="input">{15N-masses}</note>'.format(**self.params['translations'])
        else:
            self.params['translations']['15N_default_input_addon'] = '<note label="protein, modified residue mass file" type="input">no</note>'

        # modifications
        potential_mods = []
        refine_potential_mods = []
        fixed_mods = []
        self.params['translations']['Prot-N-term'] = 0.0
        self.params['translations']['Prot-C-term'] = 0.0
        for mod in self.params[ 'mods' ][ 'fix' ]:
            fixed_mods.append(
                '{0}@{1}'.format(mod[ 'mass' ], mod[ 'aa' ] )
            )

        self.params['translations']['acetyl_N_term'] = 'no'
        self.params['translations']['pyro_glu'] = 'no'
        pyro_glu = 0
        potentially_modified_aa = set()
        for mod in self.params[ 'mods' ][ 'opt' ]:
            if mod[ 'aa' ] == '*' and mod[ 'name' ] == 'Acetyl' and mod[ 'pos' ] == 'Prot-N-term' :
                self.params['translations']['acetyl_N_term'] = 'yes'
                continue
            if mod[ 'aa' ] == '*' and mod[ 'name' ] == 'Gln->pyro-Glu' and mod[ 'pos' ] == 'Pep-N-term' :
                pyro_glu += 1
                continue
            if mod[ 'aa' ] == '*' and mod[ 'name' ] == 'Glu->pyro-Glu' and mod[ 'pos' ] == 'Pep-N-term' :
                pyro_glu += 1
                continue
            for term in ['Prot-N-term', 'Prot-C-term']:
                if mod[ 'pos' ] == term:
                    if mod[ 'aa' ] == '*':
                        if self.params['translations'][term] != 0.0:
                            print(
                                '''
            [ WARNING ] X!Tandem does not allow two mods on the same position {1}
            [ WARNING ] Continue without modification {0} '''.format(mod, term, **mod)
                            )
                            continue
                        else:
                            self.params['translations'][term] = mod[ 'mass' ]
                    else:
                        print(
                                '''
            [ WARNING ] X!Tandem does not support specific aminoacids for terminal modifications
            [ WARNING ] Continue without modification {0} '''.format(mod, term, **mod)
                            )
                        continue
            if mod['aa'] in potentially_modified_aa:
                print(
                    '''
            [ WARNING ] X!Tandem does not allow two potential mods on the same aminoacid!
            [ WARNING ] Continue without modification {0} '''.format(mod, **mod)
                )
                continue
            else:
                potential_mods.append(
                    '{0}@{1}'.format(mod[ 'mass' ], mod[ 'aa' ] )
                )
                potentially_modified_aa.add( mod['aa'] )

        if pyro_glu == 2:
            self.params['translations']['pyro_glu'] = 'yes'
        if pyro_glu == 1:
            print('''
    [ WARNING ] X!Tandem looks for Gln->pyro-Glu and Glu->pyro-Glu
    [ WARNING ] at the same time, please include both or none
    [ WARNING ] Continue without modification {0} '''.format(mod, **mod)
            )
        self.params['translations']['fixed_modifications'] =  ','.join( fixed_mods )
        self.params['translations']['potential_modifications'] = ','.join( potential_mods )
        self.params['translations']['refine_potential_modifications'] = ','.join( refine_potential_mods )

        for ion in ['a', 'b', 'c', 'x', 'y', 'z']:
            if ion in self.params['translations']['score_ion_list']:
                self.params['translations']['score_{0}_ions'.format(ion)] = 'yes'
            else:
                self.params['translations']['score_{0}_ions'.format(ion)] = 'no'

        templates = self.format_templates( )
        for file_name, content in templates.items():
            if file_name == '15N-masses.xml' and self.params['translations']['label'] == '14N':
                continue
            xml_file_path = os.path.join(
                self.params['output_dir_path'],
                file_name
            )
            with open( xml_file_path, 'w') as out:
                print( content, file=out)
                self.print_info('wrote input file {0}'.format( file_name ))

                self.created_tmp_files.append( xml_file_path )
        return self.params


    def postflight( self ):
        # convert_xtandemXML_to_identcsv( self.params )
        pass

    def format_templates( self ):
        '''Returns formatted X!Tandem input files

        The formating is taken from self.params

        Returns:
            dict: keys are the names of the three templates (15N-masses.xml, taxonomy.xml, input.xml)

        '''
        templates = {
            '15N-masses.xml' : '''\
<?xml version="1.0"?>
    <bioml title="peptide residue molecular mass values for an all 15N organisms">
        <aa type="A" mass="72.034148698" />
        <aa type="B" mass="116.036998" />
        <aa type="C" mass="104.006219398" />
        <aa type="D" mass="116.023977958" />
        <aa type="E" mass="130.039628028" />
        <aa type="F" mass="148.065448838" />
        <aa type="G" mass="58.018498628" />
        <aa type="H" mass="140.050016555" />
        <aa type="I" mass="114.081098908" />
        <aa type="J" mass="0.0" />
        <aa type="K" mass="130.089032837" />
        <aa type="L" mass="114.081098908" />
        <aa type="M" mass="132.037519538" />
        <aa type="N" mass="116.036997257" />
        <aa type="O" mass="0.0" />
        <aa type="P" mass="98.049798768" />
        <aa type="Q" mass="130.052647327" />
        <aa type="R" mass="160.089250624" />
        <aa type="S" mass="88.029063328" />
        <aa type="T" mass="102.044713398" />
        <aa type="V" mass="100.065448838" />
        <aa type="W" mass="188.073382767" />
        <aa type="X" mass="112.057034" />
        <aa type="Y" mass="164.060363468" />
        <aa type="Z" mass="130.052648" />
        <molecule type="NH3" mass="18.02358311" />
        <molecule type="H2O" mass="18.01056470" />
    </bioml>
''',
            # -------------------------
            # -------------------------
            'taxonomy.xml' : \
'''<?xml version='1.0' encoding='iso-8859-1'?>
    <bioml label="x! taxon-to-file matching list">
      <taxon label="{database_taxonomy}">
       <file URL="{database}" format="peptide" />
     </taxon>
    </bioml>
'''.format(**self.params['translations']),
            # -------------------------
            # -------------------------
            'input.xml' : '''<?xml version='1.0' encoding='iso-8859-1'?>
    <bioml>
      <note label="list path, default parameters" type="input">{default_input}</note>
      <note label="list path, taxonomy information" type="input">{taxonomy}</note>
      <note label="spectrum, path" type="input">{mgf_input_file}</note>
      <note label="output, path" type="input">{output_file_incl_path}</note>
    </bioml>'''.format( **self.params['translations'] ),

        'default_input.xml' : '''<?xml version='1.0' encoding='iso-8859-1'?>
    <bioml label="ursgal">
    <note type="heading">

        Spectrum general

    </note>
    <note type="input" label="spectrum, parent monoisotopic mass error plus">{precursor_mass_tolerance_plus}</note>
    <note type="input" label="spectrum, parent monoisotopic mass error minus">{precursor_mass_tolerance_minus}</note>
    <note type="input" label="spectrum, parent monoisotopic mass error units">{precursor_mass_tolerance_unit}</note>
    <note type="input" label="spectrum, parent monoisotopic mass isotope error">{precursor_isotope_range}</note>
    <note type="input" label="spectrum, minimum parent m+h">{precursor_min_mass}</note>

    <note type="input" label="spectrum, fragment mass type">{frag_mass_type}</note>
    <note type="input" label="spectrum, fragment monoisotopic mass error">{frag_mass_tolerance}</note>
    <note type="input" label="spectrum, fragment monoisotopic mass error units">{frag_mass_tolerance_unit}</note>
    <note type="input" label="spectrum, fragment mass error">{frag_mass_tolerance}</note>
    <note type="input" label="spectrum, fragment mass error units">{frag_mass_tolerance_unit}</note>
    <note type="heading">

        Spectrum conditioning

    </note>
    <note type="input" label="spectrum, dynamic range">{spec_dynamic_range}</note>
    <note type="input" label="spectrum, total peaks">{max_accounted_observed_peaks}</note>
    <note type="input" label="spectrum, use noise suppression">{noise_suppression_enabled}</note>
    <note type="input" label="spectrum, use neutral loss window">{neutral_loss_enabled}</note>
    <note type="input" label="spectrum, neutral loss window">{neutral_loss_window}</note>
    <note type="input" label="spectrum, neutral loss mass">{neutral_loss_mass}</note>
    <note type="input" label="spectrum, minimum fragment mz">{frag_min_mz}</note>
    <note type="input" label="spectrum, minimum peaks">{min_required_observed_peaks}</note>
    <note type="input" label="spectrum, threads">{cpus}</note>
    <note type="input" label="spectrum, sequence batch size" >{batch_size}</note>
    <note type="input" label="spectrum, use noise suppression" >{noise_suppression_enabled}</note>
    <note type="heading">

        Residue modification

    </note>
    <note type="input" label="residue, modification mass">{fixed_modifications}</note>
    <note type="input" label="residue, potential modification mass">{potential_modifications}</note>
    <note type="heading">
        Protein general
    </note>
    <note type="input" label="protein, taxon">{database_taxonomy}</note>
    <note type="input" label="protein, cleavage site">{enzyme}</note>
    <note type="input" label="protein, cleavage C-terminal mass change">{cleavage_cterm_mass_change}</note>
    <note type="input" label="protein, cleavage N-terminal mass change">{cleavage_nterm_mass_change}</note>
    <note type="input" label="protein, cleavage semi">{semi_enzyme}</note>
    <note type="input" label="protein, N-terminal residue modification mass">{Prot-N-term}</note>
    <note type="input" label="protein, C-terminal residue modification mass">{Prot-C-term}</note>
    <note type="input" label="protein, quick acetyl" >{acetyl_N_term}</note>
    <note type="input" label="protein, quick pyrolidone" >{pyro_glu}</note>
    <note type="input" label="protein, saps" >{search_for_saps}</note>
    <note type="input" label="protein, stP bias" >{xtandem_stp_bias}</note>
    {15N_default_input_addon}
    <note type="heading">
        Scoring
    </note>
    <note type="input" label="scoring, a ions">{score_a_ions}</note>
    <note type="input" label="scoring, b ions">{score_b_ions}</note>
    <note type="input" label="scoring, c ions" >{score_c_ions}</note>
    <note type="input" label="scoring, minimum ion count">{min_required_matched_peaks}</note>
    <note type="input" label="scoring, maximum missed cleavage sites">{max_missed_cleavages}</note>
    <note type="input" label="scoring, cyclic permutations" >{compensate_small_fasta}</note>
    <note type="input" label="scoring, include reverse" >{engine_internal_decoy_generation}</note>
    <note type="input" label="scoring, x ions" >{score_x_ions}</note>
    <note type="input" label="scoring, y ions" >{score_y_ions}</note>
    <note type="input" label="scoring, z ions" >{score_z_ions}</note>
    <note type="heading">
        Model refinement parameters
    </note>
    <note type="input" label="refine">{use_refinement}</note>
    <note type="input" label="refine, spectrum synthesis">no</note>
    <note type="input" label="refine, maximum valid expectation value">0.1</note>
    <note type="input" label="refine, potential N-terminus modifications"></note>
    <note type="input" label="refine, unanticipated cleavage">no</note>
    <note type="input" label="refine, potential modification mass">{refine_potential_modifications}</note>
    <note type="input" label="refine, use potential modifications for full refinement">no</note>
    <note type="input" label="refine, point mutations">no</note>
    <note type="heading">
        Output
    </note>
    <note type="input" label="output, path hashing">no</note>
    <note type="input" label="output, xsl path">tandem-style.xsl</note>
    <note type="input" label="output, parameters">yes</note>
    <note type="input" label="output, performance">yes</note>
    <note type="input" label="output, spectra">yes</note>
    <note type="input" label="output, histograms">yes</note>
    <note type="input" label="output, proteins">yes</note>
    <note type="input" label="output, sequences">yes</note>
    <note type="input" label="output, results">all</note>
    <note type="input" label="output, maximum valid expectation value">{max_output_e_value}</note>
    <note type="input" label="output, histogram column width">30</note>
    </bioml>'''.format(**self.params['translations'])
            }
        return templates



#!/usr/bin/env python

from .myrimatch_2_1_138 import myrimatch_2_1_138 as myri


class myrimatch_2_2_140( myri ):
    """
    Myrimatch UNode

    Import functions from myrimatch_2_1_138
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'Myrimatch',
        'version'                     : '2.2.140',
        'release_date'                : None,
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'input_extensions'            : ['.mzML'],
        'output_extensions'           : ['.mzid'],
        'create_own_folder'           : True,
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'               : True,
        'utranslation_style'          : 'myrimatch_style_1',
        'engine' : {
            'win32' : {
                '64bit' : {
                    'exe'            : 'myrimatch.exe',
                    'url'            : '',
                    'zip_md5'        : '36e6e118ca0a8efda19b2556bd18f41e',
                    'additional_exe' : [],
                },
                # '32bit' : {
                #     'exe'            : 'myrimatch.exe',
                #     'url'            : '',
                #     'zip_md5'        : '59d00d759ae6f9225f42ed72844a9ca2',
                #     'additional_exe' : [],
                # },
            },
        },
        'citation' : \
            'Tabb DL, Fernando CG, Chambers MC. (2007) MyriMatch: highly '\
            'accurate tandem mass spectral peptide identification by '\
            'multivariate hypergeometric analysis.',
    }
    pass



#!/usr/bin/env python
'''
Convert .csv files to counted result files (.csv)

usage:
    ./csv2counted_results_1_0_0.py <input_file> <output_file>

Note:
    The input_file should be a unified, merged (if applicable) result file

'''

import sys
import os
import csv
import ursgal

if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)


def main(
    input_file=None,
    output_file=None,
    identifier_colum_names=None,  # list of fieldnames
    count_column_names=None,  # list of fieldnames
    count_by_file=True,
    convert2sfinx=False,
    keep_column_names=None,  # list of fieldnames
):
    '''
    Results (.csv) are summarized as table (.csv) containing all identified
    proteins, peptides, or other specified identifiers. For each sample,
    the peptide or spectral count for each identifier is given.

    This can be used to convert .csv files to SFINX input files.

    This is a .csv file containing unique peptide counts
    for all identified proteins. However, this can be modified
    using the keywords "identifier_colum_names" and "count_column_names"

    Keyword arguments:

        input_file (str): name including path for the input file
        output_file (str): name including path for  the output file
        identifier_colum_names (list): list of column headers
            that define the identifier. Multiple column names are
            joined for combined identifiers.
        count_column_names (list): list of column headers which
            are used for counting.
        count_by_file (bool): the number of unique hits for each
            identifier is given in seperate columns for each raw file
            (file name as defiened in Spectrum Title)
        convert2sfinx (bool): If True, the header of the identifier column
            is "rownames". If False, the joined header name will be used.
        keep_column_names (list): list of column headers which
            are not used as identifiers but kept in the output,
            e.g. when counting ['Sequence', 'Modifications'] the column
            ['Protein ID'] could be specified here. Multiple entries 
            for one identifier (e.g. when identifier_column_names = ['Potein ID']
            and keep_column_names = ['Sequence']) are seperated by '<#>'. 
    '''

    csv_kwargs = {}
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'

    output_file_object = open(output_file, 'w')

    count_dict, all_filenames = count_results(
        identifier_colum_names=identifier_colum_names,
        count_column_names=count_column_names,
        count_by_file=count_by_file,
        input_file=input_file,
        keep_column_names=keep_column_names,
    )
    if convert2sfinx:
        new_fieldnames = ['rownames']
    else:
        new_fieldnames = []
        for fieldname in identifier_colum_names:
            new_fieldnames.append(fieldname)
        if keep_column_names is not None:
            for fieldname in keep_column_names:
                new_fieldnames.append(fieldname)

    for filename in sorted(all_filenames):
        new_fieldnames.append(filename)

    csv_output = csv.DictWriter(
        output_file_object,
        new_fieldnames,
        **csv_kwargs
    )
    csv_output.writeheader()

    for identifier in count_dict.keys():
        out_dict = {}
        for i, fieldname in enumerate(identifier_colum_names):
            out_dict[fieldname] = identifier.split('<#>')[i]
        for filename in all_filenames:
            count = len(count_dict[identifier].get(filename, []))
            out_dict[filename] = count
        if keep_column_names is not None:
            for column_name in keep_column_names:
                out_dict[column_name] = '<#>'.join(
                    count_dict[identifier][column_name]
                )
        csv_output.writerow(out_dict)


def count_results(
    identifier_colum_names=None,  # list of fieldnames
    count_column_names=None,  # list of fieldnames
    count_by_file=True,
    input_file=None,
    keep_column_names=None,  # list of fieldnames
):
    count_dict = {}
    all_filenames = set(['all'])
    with open(input_file, 'r') as in_file:
        csv_input = csv.DictReader(in_file)
        for csv_line_dict in csv_input:
            identifier = []
            for column_name in identifier_colum_names:
                identifier.append(csv_line_dict[column_name])
            identifier = '<#>'.join(identifier)
            if identifier not in count_dict.keys():
                count_dict[identifier] = {'all': set()}
            countable = []
            for column_name in count_column_names:
                countable.append(csv_line_dict[column_name])
            countable = '<#>'.join(countable)
            if count_by_file:
                filename = csv_line_dict['Spectrum Title'].split('.')[0]
            else:
                filename = 'all'
            all_filenames.add(filename)
            if filename not in count_dict[identifier].keys():
                count_dict[identifier][filename] = set()
            count_dict[identifier][filename].add(countable)
            count_dict[identifier]['all'].add(countable)
            if keep_column_names is not None:
                for column_name in keep_column_names:
                    assert column_name not in all_filenames, '''
                    Column names in 'keep_column_names' cannot be equal
                    to file names.
                    conflicting column/file name: {0}
                    '''.format(column_name)
                    if column_name not in count_dict[identifier]:
                        count_dict[identifier][column_name] = set()
                    count_dict[identifier][column_name].add(
                        csv_line_dict[column_name]
                    )

    return count_dict, all_filenames

if __name__ == '__main__':
    main(
        input_file=sys.argv[1],
        output_file=sys.argv[2],
        identifier_colum_names=['Protein ID'],
        count_column_names=['Sequence'],
        count_by_file=True,
        keep_column_names=['Modifications'],
    )



#!/usr/bin/env python
import ursgal
import os


class mascot_dat2csv_1_0_0(ursgal.UNode):
    """
    Dummy to merge mascot data into usgal workflow
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'mascot_dat2csv',
        'version'                     : '1.0.0',
        'engine_type' : {
            'converter': True
        },
        'output_extensions'           : ['.csv'],
        'create_own_folder'           : False,
        'utranslation_style'          : 'mascot_style_1',
        'citation'                    : 'www.matrixscience.com',
        'in_development'              : False,
        'include_in_git'              : True,
        'input_extensions'            : ['.dat'],
        'release_date'                : None,
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'mascot_dat2csv_1_0_0.py',
                },
            },
        },
        'citation' : '',
    }

    def __init__(self, *args, **kwargs):
        super(mascot_dat2csv_1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        dat2csv = self.import_engine_as_python_function()

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        dat2csv(input_file, output_file)
        # self._add_mascot_to_ursgal_history(output_file)
        return output_file

    def _add_mascot_to_ursgal_history(self, file):
        """Add Mascot to history.

        Args:
            mapped_csv_search_results (str): Path to mapped search results.
        """
        # print('CHECKING HISTORY')
        mapped_json = '{0}.u.json'.format(file)
        mascot_missing = True
        lines = []
        with open(mapped_json, 'r') as org:
            for line in org:
                if 'mascot_x_x_x' in line:
                    mascot_missing = False
                    break
                lines.append(line)

        if mascot_missing:
            with open(mapped_json, 'w') as new:
                for line in lines:
                    if '"history": [' in line:
                        print('''
                "history": [
                  {
                    "META_INFO": {
                      "engine_type": {
                        "protein_database_search_engine": true,
                        "By Hand" : true
                      }
                    },
                    "engine": "mascot_x_x_x"
                  },''', file = new)
                    else:
                        print(line, end='', file = new)
